
import json
# from operator import and_, or_
from sqlalchemy import and_,or_

import pandas as pd
from shortuuid import uuid
from brave.microbe.schemas.entity import AddMeshNode, PageEntity
from brave.microbe.models.core import t_mesh,t_mesh_tree,t_registry_numbers,t_taxonomy
from sqlalchemy import select,func
from sqlalchemy.engine import Connection
from sqlalchemy import select,func,insert,case,exists

from brave.microbe.utils import import_ncbi_mesh
from brave.microbe.utils import import_kegg
from sqlalchemy.dialects.mysql import insert as mysql_insert



def page(conn: Connection, query: PageEntity):
    
    child = t_mesh_tree.alias("child")
    select_list = [
            t_mesh.c.entity_id,
         
            func.any_value(t_mesh.c.is_research).label("is_research"),
            func.any_value(t_mesh.c.entity_type).label("entity_type"),

            func.any_value(t_mesh.c.tags).label("tags"),
            func.any_value(t_mesh.c.short_name).label("short_name"),

            # func.any_value(t_mesh.c.describe).label("describe"),
            func.group_concat(func.distinct(t_mesh_tree.c.parent_tree)).label("parent_trees"),
            func.any_value(t_mesh_tree.c.tree_number).label("tree_number"),
            func.any_value(case(
                    (exists().where(child.c.parent_tree == t_mesh_tree.c.tree_number), True),
                    else_=False
            )).label("has_children"),
    ]
    if query.locale and query.locale=="zh_CN":
        select_list.extend([
            func.any_value(t_mesh.c.entity_name).label("entity_name_en"),
            func.any_value(
                func.coalesce(
                    func.nullif(t_mesh.c.entity_name_zh, ''),  # 如果 entity_name_zh = '' → NULL
                    t_mesh.c.entity_name                       # 如果为 NULL 或空 → 用 entity_name
                )
            ).label("entity_name"),
        ])
        # stmt =select(
           
        #     # t_mesh_tree.c.parent_tree.label("parent_tree"),
        #     # t_mesh_tree.c.tree_number.label("tree_number"),
        # )

    else:    
        select_list.extend([
            func.any_value(t_mesh.c.entity_name).label("entity_name")
        ])
    stmt =select(*select_list)
    
    # stmt = stmt.with_only_columns(select_list)

    if query.registry_join:
        if  query.registry_join=="taxonomy":
            stmt = select(*select_list,
            func.group_concat(func.distinct(t_taxonomy.c.entity_name)).label("taxonomy_name"),
            func.group_concat(func.distinct(t_taxonomy.c.rank)).label("rank"),
            func.group_concat(func.distinct(t_taxonomy.c.tax_id)).label("tax_id"),
                          )
        
            stmt = stmt.outerjoin(t_registry_numbers, t_mesh.c.entity_id == t_registry_numbers.c.entity_id)
            stmt = stmt.outerjoin(t_taxonomy, t_registry_numbers.c.registry_number == t_taxonomy.c.entity_id)


    if query.parent_id and query.parent_id=="default":
        query.parent_id = query.category
        
    stmt = stmt.select_from(
        t_mesh.outerjoin(t_mesh_tree,t_mesh.c.entity_id==t_mesh_tree.c.entity_id)
            )
  
    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
             t_mesh.c.entity_name.ilike(keyword_pattern)
        )
    if query.category:
        if type (query.category) is str:
            conditions.append(
                t_mesh_tree.c.category == query.category
            )
        elif type (query.category) is list:
            conditions.append(
                t_mesh_tree.c.category.in_(query.category)
            )
    # if query.major_category:
    #     conditions.append(
    #          t_mesh_tree.c.major_category == query.major_category
    #     )
    if query.parent_id:
        if type (query.parent_id) is str:
            conditions.append(
                t_mesh_tree.c.parent_tree == query.parent_id
            )
        elif type (query.parent_id) is list:
            conditions.append(
                t_mesh_tree.c.tree_number.in_(query.parent_id)
        
            )
    if query.is_research is not None and query.is_research is True:
        conditions.append(
             t_mesh.c.is_research == query.is_research
        )   
    
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count( func.distinct(t_mesh.c.entity_id) )).select_from(t_mesh.outerjoin(t_mesh_tree,t_mesh.c.entity_id==t_mesh_tree.c.entity_id)).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_mesh)
    
    stmt = stmt.group_by(t_mesh.c.entity_id)

    if query.page_size != -1:
        stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_disease = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    return {
        "items": find_disease,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }

def get_parents(conn, entity_id: str):
    # 获取该 entity_id 对应的所有 tree_number
    tree_numbers = conn.execute(
        select(t_mesh_tree.c.tree_number).where(t_mesh_tree.c.entity_id == entity_id)
    ).scalars().all()
    
    if not tree_numbers:
        return []
    
    # 生成所有父前缀
    parent_prefixes = []
    for tn in tree_numbers:
        parent_prefixes.extend(tree_numbers)
        parts = tn.split(".")
        parent_prefixes.extend([".".join(parts[:i]) for i in range(1, len(parts))])
    parent_prefixes = list(set(parent_prefixes))
    
    if not parent_prefixes:
        return []
    
    # 查询所有父节点
    stmt_parents = (
        select(t_mesh.c.entity_name, t_mesh_tree.c.entity_id, t_mesh_tree.c.tree_number, t_mesh_tree.c.parent_tree)
        .select_from(t_mesh_tree.outerjoin(t_mesh, t_mesh.c.entity_id == t_mesh_tree.c.entity_id))
        .where(t_mesh_tree.c.tree_number.in_(parent_prefixes))
        .order_by(t_mesh_tree.c.tree_number)
    )
    all_nodes = conn.execute(stmt_parents).mappings().all()
    if not all_nodes:
        return []

    # 4. 构建字典: tree_number -> 节点
    nodes_dict = {node['tree_number']: dict(node) for node in all_nodes}
    # 初始化 children 列表
    for node in nodes_dict.values():
        node['children'] = []

    # 5. 构建树结构
    tree_roots = []
    for tn, node in nodes_dict.items():
        parent_tn = node['parent_tree']
        if parent_tn and parent_tn in nodes_dict:
            nodes_dict[parent_tn]['children'].append(node)
        else:
            tree_roots.append(node)  # 根节点

    return tree_roots
def find_by_id(conn: Connection, entity_id: str):
    stmt = t_mesh.select().where(t_mesh.c.entity_id ==entity_id)
    find_disease = conn.execute(stmt).mappings().first()
    parents  = get_parents(conn, entity_id)
    find_disease = dict(find_disease) if find_disease else {}
    find_disease["parents"] = parents
    return find_disease



def update(conn: Connection, entity_id: str, updateData: dict):
    stmt = t_mesh.update().where(t_mesh.c.entity_id == entity_id).values(updateData)
    conn.execute(stmt)

def add(conn: Connection, data: dict):
    data['entity_id'] = str(uuid())
    stmt = t_mesh.insert().values(data)
    conn.execute(stmt)

def find_by_keywords(conn: Connection, keywords: str):
    keyword_pattern = f"%{keywords.strip()}%"
    stmt = t_mesh.select().where(t_mesh.c.entity_name.ilike(keyword_pattern)).limit(10)
    find_disease = conn.execute(stmt).mappings().all()
    return find_disease

def details_by_id(conn: Connection, entity_id: str):
    stmt = t_mesh.select().where(t_mesh.c.entity_id == entity_id)
    details_disease = conn.execute(stmt).mappings().first()
    return details_disease

# def imports(conn: Connection, records: list, batch_size: int = 1000):
#     inserted = 0

#     for i in range(0, len(records), batch_size):
#         batch = records[i:i + batch_size]
#         stmt = t_mesh.insert()
#         conn.execute(stmt, batch)
#         inserted += len(batch)
#         print(f"Inserted {inserted} records...")

#     return {"message imports mesh": f"导入完成，共导入 {inserted} 行"}




def imports(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0

    for i in range(0, len(records), batch_size):
        batch = records[i:i + batch_size]

        stmt = mysql_insert(t_mesh).values(batch)
        # 如果 entity_id 已存在，则更新其它字段
        stmt = stmt.on_duplicate_key_update(
            entity_name=stmt.inserted.entity_name,
            # second_entity_id=stmt.inserted.second_entity_id
        )

        conn.execute(stmt)
        inserted += len(batch)
        print(f"Processed {inserted} records...")

    return {"message imports mesh": f"导入完成，共处理 {inserted} 行"}



def imports_tree(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0
    for i in range(0, len(records), batch_size):
        batch = records[i:i + batch_size]

        stmt = mysql_insert(t_mesh_tree).values(batch)
        # 如果 entity_id 已存在，则更新其它字段
        stmt = stmt.on_duplicate_key_update(

            category=stmt.inserted.category
        )

        conn.execute(stmt)
        inserted += len(batch)
        print(f"Processed {inserted} records...")
    return {"message imports mesh": f"导入完成，共处理 {inserted} 行"}

def  import_registry_numbers(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0
    for i in range(0, len(records), batch_size):
        batch = records[i:i + batch_size]

        stmt = mysql_insert(t_registry_numbers).values(batch)
        # 如果 entity_id 已存在，则更新其它字段
        stmt = stmt.on_duplicate_key_update(

            registry_type=stmt.inserted.registry_type
        )

        conn.execute(stmt)
        inserted += len(batch)
        print(f"Processed {inserted} records...")
    return {"message imports registry_numbers": f"导入完成，共处理 {inserted} 行"}

def delete_by_id(conn: Connection, entity_id: str):
    stmt = t_mesh.delete().where(t_mesh.c.entity_id == entity_id)
    conn.execute(stmt)
    stmt = t_mesh_tree.delete().where(t_mesh_tree.c.entity_id==entity_id)
    conn.execute(stmt)



def get_parent(tree_number):
    if pd.isna(tree_number):
        return None
    parts = tree_number.split(".")
    return ".".join(parts[:-1]) if len(parts) > 1 else None

# def imports_tree(conn: Connection, records: list, batch_size: int = 1000):
#     inserted = 0
#     for i in range(0, len(records), batch_size):
#         batch = records[i:i + batch_size]
#         stmt = t_mesh_tree.insert()
#         conn.execute(stmt, batch)
#         inserted += len(batch)
#         print(f"Inserted {inserted} records...")

#     return {"message imports_tree": f"导入完成，共导入 {inserted} 行"}

# def imports_tree(conn: Connection, records: list, batch_size: int = 1000):
#     inserted = 0

#     for i in range(0, len(records), batch_size):
#         batch = records[i:i + batch_size]

#         stmt = insert(t_mesh_tree).values(batch)
#         stmt = stmt.on_duplicate_key_update(
#             category=stmt.inserted.category
#         )

#         conn.execute(stmt)
#         inserted += len(batch)
#         print(f"Processed {inserted} records...")

#     return {"message imports_tree": f"导入完成，共处理 {inserted} 行"}

def init(conn: Connection,category):
    if category == "mesh":
        json_data = import_ncbi_mesh.get_json("/ssd1/wy/workspace2/nextflow-fastapi/desc2025.xml")
        df = pd.DataFrame(json_data)
        df['second_entity_id'] = df['registry_numbers'].apply(
            lambda x: x[0].replace("txid","TAX") if isinstance(x, list) and len(x)>0 and x[0].startswith("txid") else ""
        )
        # df = df.drop("registry_numbers",axis=1)

        df_exploded = df.explode(["TreeNumberList"], ignore_index=True)

        df_exploded["ParentTree"] = df_exploded["TreeNumberList"].apply(get_parent)
        df_exploded_rename = df_exploded[["DescriptorUI","DescriptorName","TreeNumberList","ParentTree","second_entity_id"]].rename({
                "DescriptorUI":"entity_id",
                "DescriptorName":"entity_name",
                "TreeNumberList":"tree_number",
                "ParentTree":"parent_tree",
                "second_entity_id":"registry_number"
            },axis=1)
        df_exploded_rename["category"]  = df_exploded_rename["tree_number"].apply(lambda x: str(x).split(".")[0])
        df_exploded_rename["registry_type"]  = df_exploded_rename["registry_number"].apply(lambda x: "taxonomy" if x.startswith("TAX")  else "_NULL_")
        df_exploded_rename["parent_tree"]  = df_exploded_rename["parent_tree"].apply(lambda x: x if not pd.isna(x) else "_NULL_")
        df_exploded_rename["tree_number"]  = df_exploded_rename["tree_number"].apply(lambda x: x if not pd.isna(x) else "_NULL_")

        # df_exploded_rename["major_category"]  = df_exploded_rename["tree_number"].apply(lambda x: str(x).split(".")[0][0])
        df_exploded_rename_node = df_exploded_rename[["entity_id","entity_name"]].drop_duplicates( keep="first")
        # df_exploded_rename_node["entity_id"] = df_exploded_rename_node.apply(lambda x: uuid(),axis=1)
        data = json.loads(df_exploded_rename_node.to_json(orient="records"))
        imports(conn, data)
        df_exploded_rename_tree = df_exploded_rename[["entity_id","tree_number","parent_tree","category"]].drop_duplicates( keep="first")
        data_tree = json.loads(df_exploded_rename_tree.to_json(orient="records"))
        imports_tree(conn, data_tree)
        df_exploded_rename_registry = df_exploded_rename[["entity_id","registry_number"]].drop_duplicates( keep="first")
        df_exploded_rename_registry = df_exploded_rename_registry[df_exploded_rename_registry["registry_number"]!=""]
        data_registry =  json.loads(df_exploded_rename_registry.to_json(orient="records"))
        import_registry_numbers(conn, data_registry)

    elif category == "kegg":
        with open("/data/DATABASES/br_br08901.txt", "r", encoding="utf-8") as f:
            lines = f.readlines()
        
        data_tree = import_kegg.parse_kegg_hierarchy(lines)
        
        # df_exploded_tree = pd.DataFrame(data_tree)
        # df_exploded_tree = df_exploded_tree[["entity_id","entity_name"]]
        # data = json.loads(df_exploded_rename_node.to_json(orient="records"))
        node = [ 
                { k:v for k,v in item.items() if k  in ["entity_id","entity_name"] }
                for item in data_tree  
            ]
        tree = [ 
                { k:v for k,v in item.items() if k  in ["entity_id","tree_number","parent_tree","category"] }
                for item in data_tree  
            ]
        
        # { k:v for k,v in data_tree if k  in ["entity_id","tree_number","parent_tree","category"] }
        pass
        imports(conn, node)
        imports_tree(conn, tree)

        # # df_exploded_rename_tree = df_exploded_rename_tree.merge(df_exploded_rename_node,on="entity_name",how="left")
        # df_exploded_rename_tree = df_exploded_rename_tree.merge(df_exploded_rename_node[["entity_name","entity_id"]],on="entity_name",how="left")
        # data_tree = json.loads(df_exploded_rename_tree.to_json(orient="records"))
        # imports_tree(conn, data_tree)
    return "success"

'''
{
    "entity_name": "KEGG",
    "category": "KEGG"
}
{
    "entity_name": "Metabolism",
    "category": "KEGG",
    "parent_tree": "KEGG"
}
{
    "entity_name": "Global and overview maps",
    "category": "KEGG",
    "parent_tree": "KEGG.001"
}
{
    "entity_name": "Metabolic pathways",
    "category": "KEGG",
    "entity_id":"map01100",
    "parent_tree": "KEGG.001.001"
}

{
    "entity_name": "Biosynthesis of secondary metabolites",
    "category": "KEGG",
    "entity_id":"map01110",
    "parent_tree": "KEGG.001.001"
}
'''

def add_mesh_node(conn: Connection, data: AddMeshNode):
    """
    同时新增 t_test 和 t_test_tree 节点，并自动生成 tree_number 和 parent_tree
    """
    # entity_id: str, entity_name: str, parent_tree: str = None
    # entity_id 
    # entity_name = data.entity_name
    # parent_tree = data.parent_tree
    # category = data.category
    data_dict = dict(data)
    node_values = {k:v for k,v in data_dict.items() if k!="parent_tree" and k!="category"}
    node_values["entity_id"] = str(uuid())
    # 1️ 插入 t_test 表
    stmt_insert = insert(t_mesh).values(**node_values)
    conn.execute(stmt_insert)
    parent_tree = data.parent_tree
    category = data.category
    # 2️ 生成 tree_number
    if parent_tree is None:
        # 根节点
        tree_number = category #entity_id[:1].upper()  # 可自定义规则
        parent_tree = "_NULL_"
    else:
        # 查询当前父节点已有多少个子节点
        stmt = select(t_mesh_tree.c.tree_number).where(t_mesh_tree.c.parent_tree == parent_tree)
        result = conn.execute(stmt).scalars().all()
        count = len(result) + 1
        tree_number = f"{parent_tree}.{count:03d}"  # 格式化为三位，如 001, 002
    

    # 3️ 插入 t_test_tree 表
    stmt_tree = insert(t_mesh_tree).values(
        entity_id=node_values["entity_id"],
        tree_number=tree_number,
        parent_tree=parent_tree,
        category=category
    )
    conn.execute(stmt_tree)
    return tree_number



from operator import and_, or_

from shortuuid import uuid
from brave.microbe.schemas.entity import PageEntity
from brave.microbe.models.core import t_diet_and_food
from sqlalchemy import select,func
from sqlalchemy.engine import Connection



def page(conn: Connection, query: PageEntity):
    stmt =select(
        t_diet_and_food,
    ) 
    
    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
             t_diet_and_food.c.entity_name.ilike(keyword_pattern)
        )
    
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count()).select_from(t_diet_and_food).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_diet_and_food)
    if query.page_size != -1:
        stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_disease = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    return {
        "items": find_disease,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }

def find_by_id(conn: Connection, entity_id: str):
    stmt = t_diet_and_food.select().where(t_diet_and_food.c.entity_id ==entity_id)
    find_disease = conn.execute(stmt).mappings().first()
    return find_disease

def update(conn: Connection, entity_id: str, updateData: dict):
    stmt = t_diet_and_food.update().where(t_diet_and_food.c.entity_id == entity_id).values(updateData)
    conn.execute(stmt)

def add(conn: Connection, data: dict):
    data['entity_id'] = str(uuid())
    stmt = t_diet_and_food.insert().values(data)
    conn.execute(stmt)

def find_by_keywords(conn: Connection, keywords: str):
    keyword_pattern = f"%{keywords.strip()}%"
    stmt = t_diet_and_food.select().where(t_diet_and_food.c.entity_name.ilike(keyword_pattern)).limit(10)
    find_disease = conn.execute(stmt).mappings().all()
    return find_disease

def details_by_id(conn: Connection, entity_id: str):
    stmt = t_diet_and_food.select().where(t_diet_and_food.c.entity_id == entity_id)
    details_disease = conn.execute(stmt).mappings().first()
    return details_disease

def imports(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0
    with conn.begin():  # type: Connection
        for i in range(0, len(records), batch_size):
            batch = records[i:i + batch_size]
            stmt = t_diet_and_food.insert()
            conn.execute(stmt, batch)
            inserted += len(batch)
            print(f"Inserted {inserted} records...")

    return {"message": f"导入完成，共导入 {inserted} 行"}

def delete_by_id(conn: Connection, entity_id: str):
    stmt = t_diet_and_food.delete().where(t_diet_and_food.c.entity_id == entity_id)
    conn.execute(stmt)


from .uds_stream import UDSStreamIngress
from .uds_socket import UDSSocketIngress
from .http_ingress import HTTPIngress
from brave.api.core.routers.ingress_event_router import IngressEventRouter

# brave/api/core/ingress_factory.py
# class IngressFactory:
#     def __init__(self, event_mode: str, uds_path: str,ingress_event_router: IngressEventRouter):
#         self.event_mode = event_mode
#         self.uds_path = uds_path
#         self.ingress_event_router = ingress_event_router
        
#     def create(self):
#         if self.event_mode == "stream":
#             return UDSStreamIngress(self.uds_path, self.ingress_event_router)
#         elif self.event_mode == "socket":
#             return UDSSocketIngress(self.uds_path, self.ingress_event_router)
#         elif self.event_mode == "http":
#             return HTTPIngress(self.ingress_event_router)
#         else:
#             raise ValueError(f"Unsupported ingress mode: {self.event_mode}")

from enum import Enum

class IngressMode(Enum):
    STREAM = "stream"
    SOCKET = "socket"
    HTTP = "http"

def create_ingress(mode: IngressMode, path, router: IngressEventRouter):
    if mode == IngressMode.STREAM:
        return UDSStreamIngress(path, router)
    elif mode == IngressMode.SOCKET:
        return UDSSocketIngress(path, router)
    elif mode == IngressMode.HTTP:
        return HTTPIngress(router)
    else:
        raise ValueError(f"Unsupported ingress mode: {mode}")


import glob
import os
from fastapi import APIRouter, HTTPException, Query
import asyncio
from fastapi.responses import FileResponse
from brave.api.schemas.file_operation import WriteFile
from collections import defaultdict
from pathlib import Path
from brave.api.config.config import get_settings
import pandas as pd
import json
import brave.api.service.file_operation as file_operation_service
import aiofiles

file_locks = defaultdict(asyncio.Lock)

file_operation = APIRouter()


@file_operation.get("/file-operation/read-file")
async def read_file(file_path):
    if not os.path.exists(file_path):
        return f"{file_path}文件不存在"
    try:
        # if file_path.endswith(".xlsx"):
        #     return pd.read_excel(file_path).to_json(orient="records")

        # with open(file_path, 'r') as file:
        #     return file.read()
        return file_operation_service.format_table_output(file_path)
    except Exception as e:
        return f"{file_path}文件读取失败: {e}"

# @file_operation.get("/file-operation/read-log-file")
# async def read_log_file(file_path,offset:int=0):
#     lock = file_locks[file_path]
#     async with lock:
#         if not os.path.exists(file_path):
#             return {
#                 "content": [],
#                 "offset": 0
#             }
#         with open(file_path, 'r') as file:
#             file.seek(offset)
#             return {
#                 "content": file.readlines(),
#                 "offset": file.tell()
#             }

@file_operation.get("/file-operation/read-log-file")
async def read_log_file(file_path: str, offset: int = 0):
    lock = file_locks[file_path]
    async with lock:
        if not os.path.exists(file_path):
            return {"content": [], "offset": 0}
        
        async with aiofiles.open(file_path, 'r') as file:
            await file.seek(offset)   # aiofiles 支持
            content = await file.readlines()
            offset = await file.tell()
        
        return {"content": content, "offset": offset}

# @app.get("/logs/delta")
# def get_incremental_logs(offset: int):
#     with open(LOG_FILE, "r") as f:
#         f.seek(offset)
#         new_data = f.read()
#         new_offset = f.tell()
#     return {
#         "logs": new_data,
#         "offset": new_offset
#     }


@file_operation.post("/file-operation/write-file")
async def write_file(writeFile:WriteFile):
    with open(writeFile.file_path, 'w') as file:
        file.write(writeFile.content)


@file_operation.get("/file-operation/file-list-recursive")
async def get_all_files_recursive(directory):
    file_list=[]
    for root, dirs, files in os.walk(directory):
        for file in files:
            file_list.append(os.path.join(root, file).replace(directory,""))
    return file_list


@file_operation.get("/file-operation/list-dir")
def list_dir(path: str = ""):
    full_path = Path( path).resolve()
    get_setting = get_settings()
    if not full_path.exists() or not full_path.is_dir():
        raise HTTPException(status_code=400, detail="Invalid path")
    
    items = []
    for item in full_path.iterdir():
        items.append({
            "name": item.name,
            "is_dir": item.is_dir(),
            "size": item.stat().st_size if item.is_file() else None,
            "modified": item.stat().st_mtime
        })
    return items


@file_operation.get("/file-operation/download")
def download_file(path: str):
    file_path = Path(path).resolve()
    get_setting = get_settings()
    if not file_path.exists() or not file_path.is_file():
        raise HTTPException(status_code=404, detail="File not found")
    return FileResponse(file_path, filename=file_path.name)



@file_operation.get("/file-operation/list-dir-v2")
def list_dir_v2(
    path: str = "",
    keyword: str = Query("", description="模糊搜索关键字"),
    page: int = 1,
    limit: int = 20,
):
    full_path = Path(  path).resolve()
    if not full_path.exists() or not full_path.is_dir() :
        raise HTTPException(status_code=400, detail="Invalid path")

    items = []
    for item in full_path.iterdir():
        if keyword.lower() in item.name.lower():
            items.append({
                "name": item.name,
                "is_dir": item.is_dir(),
                "size": item.stat().st_size if item.is_file() else None,
                "modified": item.stat().st_mtime,
            })

    # 分页处理
    total = len(items)
    start = (page - 1) * limit
    end = start + limit
    paged_items = items[start:end]

    return {
        "items": paged_items,
        "total": total,
        "page": page,
        "limit": limit,
    }




# def format_img_path(path):
#     settings = get_settings()
#     base_dir = settings.BASE_DIR
#     file_name = path.replace(str(base_dir),"")
#     # img_base64 = base64.b64encode(open(path, 'rb').read()).decode('utf-8')
#     return {
#         "data":f"/brave-api/dir{file_name}",
#         "type":"img",
#         "url":f"/brave-api/dir{file_name}"
#     }

# def format_table_output(path):
#     # pd.set_option("display.max_rows", 1000)     # 最多显示 1000 行
#     # pd.set_option("display.max_columns", 500)   # 最多显示 500 列
#     data = ""
#     data_type="table"
#     if path.endswith("xlsx"):
#         df = pd.read_excel(path, nrows=100).iloc[:, :50]
#         data = json.loads(df.to_json(orient="records")) 
#         data_type="table"
#     elif path.endswith("txt"):
#         with open(path,"r") as f:
#             data = f.read()
#         data_type="string"
#     else:
#         df = pd.read_csv(path,sep="\t", nrows=100).iloc[:, :50]
#         # df = pd.read_csv(path,sep="\t")
#         data = json.loads(df.to_json(orient="records")) 
#         data_type="table"

#     settings = get_settings()
#     base_dir = settings.BASE_DIR
#     file_name = path.replace(str(base_dir),"")
#     return  {
#         "data":data ,
#         "type":data_type,
#         "url":f"/brave-api/dir{file_name}"
#     }
# def format_table_output(path):
#     with open(path,"r") as f:
#         text = f.read()
#     settings = get_settings()
#     base_dir = settings.BASE_DIR
#     file_name = path.replace(str(base_dir),"")
#     return  {
#         "data":text ,
#         "type":"table",
#         "url":f"/brave-api/dir{file_name}"
#     }
@file_operation.get("/file-operation/visualization-results")
async def visualization_results(path):
    return await file_operation_service.visualization_results(path)
    # path = f"{path}/output"
    # images = []
    # for ext in ("*.png", "*.jpg", "*.jpeg"):
    #     images.extend(glob.glob(os.path.join(path, ext)))
    # images = [format_img_path(image) for image in images]
    # tables = []
    # for ext in ("*.csv", "*.tsv","*.txt", "*.xlsx"):
    #     tables.extend(glob.glob(os.path.join(path, ext)))
    # tables = [format_table_output(table) for table in tables]

    # # textList = []
    # # for ext in ("*.txt"):
    # #     textList.extend(glob.glob(os.path.join(path, ext)))

    # # textList = [format_text_output(text) for text in textList]

    # return {
    #     "images": images,
    #     "tables": tables
    # }




from brave.api.executor.base import JobExecutor
from brave.api.schemas.container import ListContainerQuery, PageContainerQuery,SaveContainer
from brave.api.models.core import t_container
from sqlalchemy import delete, select, and_, join, func,insert,update
# from brave.api.service.pipeline import get_pipeline_dir 
import json

from brave.api.config.config import get_settings

def page_container(conn,query:PageContainerQuery):
    stmt =select(t_container) 
    
    stmt =select(
        t_container
    ) 
    
   
    conditions = []
 
    stmt = stmt.where(and_(*conditions))
    count_stmt = select(func.count()).select_from(t_container).where(and_(*conditions))

    stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_container = conn.execute(stmt).mappings().all()
    find_container = [dict(item) for item in find_container]
    # find_container = [format_pipeline_componnet_one() for item in find_container]

    total = conn.execute(count_stmt).scalar()
    return {
        "items": find_container,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }





def find_container_by_id(conn,container_id):
    stmt = select(
        t_container
    )
   
    stmt = stmt.where(t_container.c.container_id ==container_id)
    find_container = conn.execute(stmt).mappings().first()
    return find_container

def save_container(conn,saveContainer):
    # str_uuid = str(uuid.uuid4())     
    # saveContainer["container_id"] = str_uuid
    stmt = t_container.insert().values(saveContainer)
    conn.execute(stmt)


def update_container(conn,container_id,updateContainer):
    stmt = t_container.update().where(t_container.c.container_id == container_id).values(updateContainer)
    conn.execute(stmt)

def delete_container(conn,container_id):
    stmt = t_container.delete().where(t_container.c.container_id == container_id)
    conn.execute(stmt)

def list_container(conn): 
    stmt = t_container.select()
    return conn.execute(stmt).mappings().all()
def get_pipeline_dir():
    settings = get_settings()
    return settings.PIPELINE_DIR

def import_container(conn,path,force=False):
    # pipeline_dir = get_pipeline_dir()
    with open(f"{path}/container.json","r") as f:
        find_container_list = json.load(f)
    for item in find_container_list:
        find_container = find_container_by_id(conn,item['container_id'])
        if find_container:
            if force:
                update_stmt = update(t_container).where(t_container.c.container_id == item['container_id']).values(item)
                conn.execute(update_stmt)
        else:
            conn.execute(insert(t_container).values(item))   
        



def list_container_key(conn,query:ListContainerQuery):
    stmt = t_container.select().where(
        and_(
            t_container.c.container_key.in_(query.container_key)
        )
        
    )
    return conn.execute(stmt).mappings().all()

def find_container_key(conn,query:ListContainerQuery):
    stmt = t_container.select().where(t_container.c.container_key==query.container_key)
    return conn.execute(stmt).mappings().first()
# def update_images_status(conn,job_executor:JobExecutor ):
#     container_list = list_container()
#     for item in container_list:
#         image = job_executor.get_image(item.image)
#         pass



def find_by_container_ids(conn,container_ids):
    if not container_ids:
        return []
    stmt = t_container.select().where(t_container.c.container_id.in_(container_ids))
    return conn.execute(stmt).mappings().all()





import glob
import os
import json


# def get_json(file):
#     return json.dumps({
#         "fastq1":file,
#         "fastq2":file.replace("_host_removed_R1.fastq.gz","_host_removed_R2.fastq.gz")
#     })

def parse(dir_path,analysis):
    # file_list = glob.glob(f"{dir_path}/*_host_removed_R1.fastq.gz")
    # # /ssd1/wy/workspace2/test/test_workspace/result/V1.0/metawrap_assembly/test_s1/final_assembly.fasta
    # # sample_name,software,content_type,content
    # result_data = [(os.path.basename(file).replace("_host_removed_R1.fastq.gz",""),"samtools","json",get_json(file)) for file in file_list]
    return []



from pydantic import BaseModel

class AddProject(BaseModel):
    project_name: str
    metadata_form: str

class UpdateProject(BaseModel):
    project_id: str
    project_name: str
    metadata_form: str


import glob
import os
from fastapi import APIRouter, HTTPException, Query
import asyncio
from fastapi.responses import FileResponse
from brave.api.schemas.file_operation import WriteFile
from collections import defaultdict
from pathlib import Path
from brave.api.config.config import get_settings
import pandas as pd
import json
import fitz  # PyMuPDF
import base64
from concurrent.futures import ThreadPoolExecutor
import threading

# def pdf_page_to_base64(pdf_path, page_number=0, zoom=2):
#     """
#     读取 PDF 指定页并转为 base64 图片
#     :param pdf_path: PDF 文件路径
#     :param page_number: 页码 (从0开始)
#     :param zoom: 缩放倍率
#     :return: base64 字符串 (PNG)
#     """
#     doc = fitz.open(pdf_path)
#     page = doc.load_page(page_number)  # 页码从 0 开始
#     mat = fitz.Matrix(zoom, zoom)      # 放大，提高清晰度
#     pix = page.get_pixmap(matrix=mat)
#     img_bytes = pix.tobytes("png")

#     # 转 base64
#     base64_str = base64.b64encode(img_bytes).decode("utf-8")
#     return base64_str

def pdf_page_to_base64(pdf_path, page_number=0, zoom=2):
    try:
        doc = fitz.open(pdf_path)
        if page_number >= len(doc):
            raise ValueError(f"PDF 只有 {len(doc)} 页，无法加载第 {page_number} 页")

        page = doc.load_page(page_number)
        mat = fitz.Matrix(zoom, zoom)
        # pix = page.get_pixmap(matrix=mat)
        pix = page.get_pixmap(matrix=mat, alpha=False, clip=fitz.Rect(0, 0, 2000, 2000))

        img_bytes = pix.tobytes("png")
        return base64.b64encode(img_bytes).decode("utf-8")

    except Exception as e:
        # 任何错误返回 None 或占位图
        print(f"[ERROR] PDF 转图片失败: {pdf_path}, {e}")
        return None


def format_img_path(path):
    # print(f"Processing {path} in thread: {threading.current_thread().name}")
    settings = get_settings()
    base_dir = settings.BASE_DIR
    file_name = path.replace(str(base_dir),"")
    img_data = f"/brave-api/dir{file_name}"
    if path.endswith("pdf"):
        # pdf_file = "example.pdf"
        b64 = pdf_page_to_base64(path, page_number=0, zoom=2)
        img_data = f"data:image/png;base64,{b64}"
    # print("data:image/png;base64," + b64[:200] + "...")  # 打印前200字符
    # img_base64 = base64.b64encode(open(path, 'rb').read()).decode('utf-8')
    return {
        "data":img_data,
        "type":"img",
        "filename":os.path.basename(path),
        "url":f"/brave-api/dir{file_name}"
    }

def format_table_output(path):
    # pd.set_option("display.max_rows", 1000)     # 最多显示 1000 行
    # pd.set_option("display.max_columns", 500)   # 最多显示 500 列
    data = ""
    data_type="table"
    order = 0
    if path.endswith("xlsx"):
        df =  pd.read_excel(path, nrows=50).iloc[:, :30] 
        data = json.loads(df.to_json(orient="records")) 
        data_type="table"
    elif path.endswith("html") :
        # with open(path,"r") as f:
        #     data = f.read()
        data_type="html"
    elif path.endswith("sh") :
        with open(path,"r") as f:
            data = f.read()
        data_type="text"
    elif path.endswith(".download.tsv"):
        # df =  pd.read_csv(path,sep="\t", nrows=50).iloc[:, :30]
        # # df = pd.read_csv(path,sep="\t")
        # data = json.loads(df.to_json(orient="records")) 
        data=[]
        data_type="download"
    elif path.endswith("tsv"):
        df =  pd.read_csv(path,sep="\t", nrows=50).iloc[:, :30]
        # df = pd.read_csv(path,sep="\t")
        data = json.loads(df.to_json(orient="records")) 
        data_type="table"
    elif path.endswith(".vis"):
        data_type = os.path.basename(path).replace(".vis","")
        # df =  pd.read_csv(path,sep="\t", dtype={"pathwayId": str})
        # df = pd.read_csv(path,sep="\t")
        with open(path) as f:
            data = json.load(f)
        # data = json.loads(df.to_json(orient="records")) 
        # data_type="kegg_map"
    elif path.endswith("json"):
        with open(path,"r") as f:
            data = f.read()
        data_type="json"
    elif path.endswith("info"):
        with open(path,"r") as f:
            data = f.read()
        data_type="info"
        order=10
    else:
        with open(path,"r") as f:
            data = f.read()
        data_type="string"
      

    settings = get_settings()
    base_dir = settings.BASE_DIR
    file_name = path.replace(str(base_dir),"")
    return  {
        "data":data ,
        "order":order,
        "type":data_type,
        "filename":os.path.basename(path),
        "url":f"/brave-api/dir{file_name}"
    }
# def format_table_output(path):
#     with open(path,"r") as f:
#         text = f.read()
#     settings = get_settings()
#     base_dir = settings.BASE_DIR
#     file_name = path.replace(str(base_dir),"")
#     return  {
#         "data":text ,
#         "type":"table",
#         "url":f"/brave-api/dir{file_name}"
#     }
async def visualization_results(path):

    path = f"{path}/output"
    images = []
    for ext in ("*.png", "*.jpg", "*.jpeg","*.pdf"):
        images.extend(glob.glob(os.path.join(path, ext)))
    # images = [format_img_path(image) for image in images]
       # 多线程处理
    # with ThreadPoolExecutor(max_workers=8) as executor:
    #     images = list(executor.map(format_img_path, images))
    # print(f"Processing visualization_results in thread: {threading.current_thread().name}")

    tasks = [asyncio.to_thread(format_img_path, img) for img in images]
    images = await asyncio.gather(*tasks)
    # images = []
    tables = []
    for ext in ("*.csv", "*.tsv","*.txt", "*.xlsx","*.info","*.vis"):
        tables.extend(glob.glob(os.path.join(path, ext)))
    tables = [format_table_output(table) for table in tables]
    tables = sorted(tables, key=lambda x: x.get("order", 0), reverse=True)
    # textList = []
    # for ext in ("*.txt"):
    #     textList.extend(glob.glob(os.path.join(path, ext)))

    # textList = [format_text_output(text) for text in textList]

    return {
        "images": images,
        "tables": tables
    }




import json
from dependency_injector.wiring import inject
from brave.api.core.evenet_bus import EventBus
from brave.api.core.routers.analysis_executer_router import AnalysisExecutorRouter
from dependency_injector.wiring import inject, Provide
from brave.api.core.routers.analysis_result_router import AnalysisResultRouter
from brave.api.core.routers_name import RoutersName
from brave.api.executor.base import JobExecutor
from brave.api.schemas.analysis import Analysis, AnalysisExecuterModal
from brave.api.schemas.analysis_result import AnalysisResult, AnalysisResultParseModal
from brave.app_container import AppContainer
from brave.api.core.event import AnalysisResultEvent, WorkflowEvent
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.executor.models import JobSpec, LocalJobSpec
from brave.api.service.sse_service import SSESessionService

@inject
def setup_handlers(
    evenet_bus:EventBus  = Provide[AppContainer.event_bus],
    sse_service:SSESessionService = Provide[AppContainer.sse_service],
    router:AnalysisResultRouter  = Provide[AppContainer.analysis_result_router]):
    
    evenet_bus.register_router(RoutersName.ANALYSIS_RESULT_ROUTER,router)

    @router.on_event(AnalysisResultEvent.ON_ANALYSIS_RESULT_ADD)
    async def on_analysis_result_add(analysis:Analysis,analysis_result:AnalysisResultParseModal):
        print(f"🚀 [on_analysis_result_add] {analysis.analysis_id}")
        data = json.dumps({
            "msg":f"分析{analysis.analysis_id}，文件{analysis_result.file_name}保存成功!", 
            "component_id":analysis_result.component_id,
            "msgType":"analysis_result"
        })
        msg = {"group": "default", "data": data}
        await sse_service.push_message(msg)
  

    @router.on_event(AnalysisResultEvent.ON_ANALYSIS_RESULT_UPDATE)
    async def on_analysis_result_update(analysis:Analysis,analysis_result:AnalysisResultParseModal):
        print(f"🚀 [on_analysis_result_update] {analysis.analysis_id}")
        data = json.dumps({
            "msg":f"分析{analysis.analysis_id}，文件{analysis_result.file_name}更新成功!", 
            "component_id":analysis_result.component_id,
            "msgType":"analysis_result"
        })
        msg = {"group": "default", "data": data}
        await sse_service.push_message(msg)



import nbformat as nbf


async def generate_notebook():
    nb = nbf.v4.new_notebook()

    cells = [
        nbf.v4.new_markdown_cell("# 自动生成的 Notebook\n这里写一些说明"),
        nbf.v4.new_code_cell("import sys\nprint(sys.argv)"),
        nbf.v4.new_code_cell("print('Hello from generated notebook!')"),
    ]

    nb['cells'] = cells

    with open("generated_notebook.ipynb", "w", encoding="utf-8") as f:
        nbf.write(nb, f)



import os
import asyncio
import socket
import json
from brave.api.core.event import IngressEvent
from brave.api.core.routers.ingress_event_router import IngressEventRouter
from .interfaces.base_ingress import BaseMessageIngress

class UDSSocketIngress(BaseMessageIngress):
    def __init__(self, path, router:IngressEventRouter):
        self.path = path
        self.router = router

    async def start(self):
        if os.path.exists(self.path):
            os.remove(self.path)
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(self.path)
        sock.listen(100)
        sock.setblocking(False)
        print(f"[UDS-SOCKET] Listening at {self.path}")
        loop = asyncio.get_running_loop()
        while True:
            client, _ = await loop.sock_accept(sock)
            asyncio.create_task(self._handle(client))

    async def _handle(self, client):
        loop = asyncio.get_running_loop()
        with client:
            while True:
                try:
                    data = await loop.sock_recv(client, 4096)
                    if not data:
                        break
                    msg = json.loads(data.decode())
                    evnet_str = msg.get("ingress_event")
                    event = IngressEvent(evnet_str)
                    if not event:
                        event = IngressEvent.NEXTFLOW_EXECUTOR_EVENT
                        print(f"[UDS-SOCKET] Unknown event type '{event}'", msg)
                        return
                    data = msg.get("data")
                    await self.router.dispatch(event,data)
                except Exception as e:
                    print(f"[UDS-SOCKET] Error: {e}")
                    break



from brave.api.config.config import  get_graph
from brave.microbe.schemas.entity import GraphQuery, NodeQuery
from py2neo import Graph, Node, Relationship

def find_entity( entity_type, entity_id: str):
    graph = get_graph()
    node = graph.nodes.match(entity_type, entity_id=entity_id).first()
    return node 





def check_nodes_exist_batch(entity_ids: list[str], label: str = None):
    graph = get_graph()
    query = f"""
    MATCH (n{':' + label if label else ''})
    WHERE n.entity_id IN $entity_ids
    RETURN n.entity_id AS entity_id
    """
    result = graph.run(query, entity_ids=entity_ids).data()
    found_ids = {r["entity_id"] for r in result}
    return {eid: eid in found_ids for eid in entity_ids}


def delete_node_by_entity_id(label: str , entity_id: str):
    graph = get_graph()
    """
    根据 entity_id 删除节点及其所有关系
    """
    query = f"""
    MATCH (n{':' + label if label else ''} {{entity_id: $entity_id}})
    OPTIONAL MATCH (n)-[r]-()
    DELETE r, n
    """
    graph.run(query, entity_id=entity_id)
    print(f"Node {entity_id} and its relationships deleted successfully")


def delete_by_node_id(node_id: int):
    graph = get_graph()
    """
    根据节点的内部 ID 删除节点及其所有关系
    """
    query = """
    MATCH (n)
    WHERE id(n) = $node_id
    DETACH DELETE n
    """
    graph.run(query, node_id=node_id)
    print(f"Node with ID {node_id} and its relationships deleted successfully")


def  get_association_details(association_id: str):
    graph = get_graph()
    query = """
    MATCH (s:study)<-[:EVIDENCED_BY]-(a:association)
    MATCH (a)-[:OBSERVED_IN]->(d:disease)
    MATCH (a)-[:SUBJECT]->(i:diet_and_food)
    MATCH (a)-[:OBJECT]->(t:taxonomy)
    WHERE a.entity_id = $association_id
    RETURN a.entity_id AS association_id, a.effect AS effect,
            collect(DISTINCT {id: s.entity_id, name: s.entity_name}) AS studies,
            collect(DISTINCT  {id: d.entity_id, name: d.entity_name})  as  diseases,
            collect(DISTINCT  {id: i.entity_id, name: i.entity_name}) as  diet_and_foods,
            collect(DISTINCT  {id: t.entity_id, name: t.entity_name}) as taxonomies

    """
    result = graph.run(query, association_id=association_id).data()
    if result:
        return result[0]
    raise ValueError(f"Association {association_id} not found")

def get_nodes_by_label(label: str, page: int = 1, page_size: int = 20):
    graph = get_graph()
    """
    分页查询指定 label 的节点
    :param label: 节点 label，例如 'taxonomy'
    :param page: 页码，从 1 开始
    :param page_size: 每页条数
    :return: 节点列表
    """
    skip = (page - 1) * page_size
    query = f"""
    MATCH (n:{label})
    RETURN n
    SKIP $skip
    LIMIT $limit
    """
    result = graph.run(query, skip=skip, limit=page_size)
    return [record["n"] for record in result]


def query_nodes(query: NodeQuery):
    graph = get_graph()
    skip = (query.page - 1) * query.page_size
    # keyword 过滤条件
    if query.keyword:
        cypher = f"""
        MATCH (n:{query.label})
        WHERE n.entity_name CONTAINS $keyword
        RETURN n
        SKIP $skip
        LIMIT $limit
        """
        params = {"keyword": query.keyword, "skip": skip, "limit": query.page_size}
    else:
        cypher = f"""
        MATCH (n:{query.label})
        RETURN n
        SKIP $skip
        LIMIT $limit
        """
        params = {"skip": skip, "limit": query.page_size}

    result = graph.run(cypher, **params)
    return [record["n"] for record in result]


def build_cypher(graphQuery:GraphQuery):
    label = graphQuery.label
    keyword = graphQuery.keyword
    entity_id = graphQuery.entity_id

  
    node_var_map = {
        "association": "a",
        "study": "s",
        "taxonomy": "t",
        "disease": "d",
        "diet_and_food": "i"
    }
 

    query = """
    MATCH (s:study)<-[:EVIDENCED_BY]-(a:association)
    """
    if "disease" in graphQuery.nodes:
        query += " MATCH (a)-[:OBSERVED_IN]->(d:disease)"
    if "diet_and_food" in graphQuery.nodes:
        query += " MATCH (a)-[:SUBJECT]->(i:diet_and_food)"
    if "taxonomy" in graphQuery.nodes:
        query += " MATCH (a)-[:OBJECT]->(t:taxonomy)"

    params = {}
    # query +=" WHERE 1=1"
   
    where_clauses = []
    if keyword:
        params["keyword"] = keyword
        keyword_filters = []
        if "taxonomy" in graphQuery.nodes:
            keyword_filters.append("t.entity_name CONTAINS $keyword")
        if "disease" in graphQuery.nodes:
            keyword_filters.append("d.entity_name CONTAINS $keyword")
        if "diet_and_food" in graphQuery.nodes:
            keyword_filters.append("i.entity_name CONTAINS $keyword")
        if "study" in graphQuery.nodes:
            keyword_filters.append("s.entity_name CONTAINS $keyword")
        if keyword_filters:
            where_clauses.append("(" + " OR ".join(keyword_filters) + ")")

    if entity_id:
        params["entity_id"] = entity_id
        # 统一 entity_id 匹配，不管节点多少
        entity_filters = []
        entity_filters.append("a.entity_id = $entity_id")
        if "taxonomy" in graphQuery.nodes:
            entity_filters.append("t.entity_id = $entity_id")
        if "disease" in graphQuery.nodes:
            entity_filters.append("d.entity_id = $entity_id")
        if "diet_and_food" in graphQuery.nodes:
            entity_filters.append("i.entity_id = $entity_id")
        
        if entity_filters:
            where_clauses.append("(" + " OR ".join(entity_filters) + ")")

    nodes_dict_where_clauses = []
    if graphQuery.nodes_dict and len(graphQuery.nodes_dict)>0:
        for node_label, ids in graphQuery.nodes_dict.items():
            if not ids and len(ids)==0:
                continue
            params[f"{node_label}_ids"] = ids
            # 节点 label 映射到变量
            
            var = node_var_map.get(node_label)
            if var:
                nodes_dict_where_clauses.append(f"{var}.entity_id IN ${node_label}_ids")
        if len(nodes_dict_where_clauses) >0:
            if graphQuery.nodes_dict_condition =='OR':
                where_clauses.append("(" + " OR ".join(nodes_dict_where_clauses) + ")")
            else:
                where_clauses.append("(" + " AND ".join(nodes_dict_where_clauses) + ")")

    if where_clauses:
        query += " WHERE " + " AND ".join(where_clauses)


    # with_nodes = ["a","s"] + [node_var_map[n] for n in graphQuery.nodes if n != "association"]
    # with_clause = "WITH " + ", ".join(with_nodes)
    # query += "\n" + with_clause

    # query += """
    # RETURN a.entity_id AS association_id, a.effect AS effect,
           

    # """   
  

#   t.entity_id AS taxonomy_id, t.entity_name AS taxonomy_name,
#             d.entity_id AS disease_id, d.entity_name AS disease_name,
#            i.entity_id AS intervention_id, i.entity_name AS intervention_name,
  
    # if "disease" in graphQuery.nodes:
    #     query += ", d.entity_id AS disease_id, d.entity_name AS disease_name"
    # if "diet_and_food" in graphQuery.nodes:
    #     query += ", i.entity_id AS diet_and_food_id, i.entity_name AS diet_and_food_name"
    # if "taxonomy" in graphQuery.nodes:
    #     query += """
    #     , t.entity_id AS taxonomy_id, t.entity_name AS taxonomy_name, 
    #         SIZE([ (t)<-[:OBJECT]-(:association) | 1 ]) AS taxonomy_links
    #     """
    # if "study" in graphQuery.nodes:
    #     query += ", s.entity_id AS study_id, s.entity_name AS study_name"
 
    return query, params,node_var_map

def build_return_cypher(query:str,graphQuery:GraphQuery):
    return_fields_map = {
        "association": ["a.entity_id AS association_id", "a.effect AS effect"],
        "study": ["s.entity_id AS study_id", "s.entity_name AS study_name"],
        "taxonomy": [
            "t.entity_id AS taxonomy_id",
            "t.entity_name AS taxonomy_name"
        ],
        "disease": ["d.entity_id AS disease_id", "d.entity_name AS disease_name"],
        "diet_and_food": ["i.entity_id AS diet_and_food_id", "i.entity_name AS diet_and_food_name"]
    }
    if graphQuery.collect_association_study:
        return_fields_map.get("association").append(
            "collect(DISTINCT {id: s.entity_id, name: s.entity_name}) AS studies"
        )
        # "collect(DISTINCT {id: s.entity_id, name: s.entity_name}) AS studies "
    if graphQuery.order_by and graphQuery.order_by in return_fields_map:
        order_by = graphQuery.order_by
        field = return_fields_map.get(order_by)
        if graphQuery.order_metric == "study":
            field.append(f"SIZE([ (t)<-[]-(:association)-[:EVIDENCED_BY]->(:study) | 1 ]) AS {order_by}_links")
        elif graphQuery.order_metric=="link":
            field.append(
                f"SIZE([ (t)<-[]-(:association) | 1 ]) AS {order_by}_links"
            )
        else:
            field.append(
                f" COUNT(DISTINCT s) AS {order_by}_links"
            )
            

    return_fields = []
    return_fields.extend(return_fields_map.get("association", []))
    for n in graphQuery.nodes:
        return_fields.extend(return_fields_map.get(n, []))
    
    query += "\nRETURN " + ", ".join(return_fields)
    return query

def build_order_cypher(query:str,graphQuery:GraphQuery):
    if graphQuery.order_by in graphQuery.nodes:
        query += f"\nORDER BY {graphQuery.order_by}_links DESC"
    return query


def build_collect_return_cypher(query:str,entity_type:str,graphQuery:GraphQuery,node_var_map:dict):
    study_var =  ["id: s.entity_id", "name: s.entity_name","effect:a.effect"]
    for n in graphQuery.nodes:
        if n == "study":
            continue
        var = node_var_map.get(n)
        if var:
            study_var.append(f"{n}_id: {var}.entity_id")
            study_var.append(f"{n}_name: {var}.entity_name")
    study_var_str = " , ".join(study_var)
    return_collect_fields_map = {
        # "association": ["collect(DISTINCT {id: a.entity_id, name: a.entity_name}) AS associations"],
        "study": ["collect(DISTINCT { "+ study_var_str +" }) AS studies"],
        "taxonomy": ["collect(DISTINCT {id: t.entity_id, name: t.entity_name}) AS taxonomys"],
        "disease": ["collect(DISTINCT {id: t.entity_id, name: t.entity_name}) AS diseases"],
        "diet_and_food": ["collect(DISTINCT {id: t.entity_id, name: t.entity_name}) AS diet_and_foods"]
    }
    return_fields_map = {
        "association": ["a.entity_id AS association_id", "a.effect AS effect"],
        "study": ["s.entity_id AS study_id", "s.entity_name AS study_name"],
        "taxonomy": [
            "t.entity_id AS taxonomy_id",
            "t.entity_name AS taxonomy_name"
        ],
        "disease": ["d.entity_id AS disease_id", "d.entity_name AS disease_name"],
        "diet_and_food": ["i.entity_id AS diet_and_food_id", "i.entity_name AS diet_and_food_name"]
    }

    return_fields = []
    # return_fields.extend(return_collect_fields_map.get("association", []))
    for n in graphQuery.nodes:
        if n != entity_type:
            return_fields.extend(return_collect_fields_map.get(n, []))

    return_fields.extend(return_fields_map.get(entity_type, []))

    query += "\nRETURN " + ", ".join(return_fields)
    return query
def find_associations_by_entity_id(entity_type:str,entity_id: str,nodes :list[str]):
    graphQuery = GraphQuery(
        entity_id=entity_id,
        nodes=nodes,
        order_by=None
    )
    query,params,node_var_map =build_cypher(graphQuery)
    query = build_collect_return_cypher(query,entity_type,graphQuery,node_var_map)
    graph = get_graph()
    result = graph.run(query, **params)
    
    # print(find_cypher)
    return result.data()

def update_entity(node, entity_dict: dict):
    graph = get_graph()
    update_fields = ["entity_name","entity_name_zh"]
    entity_dict = {k: v for k, v in entity_dict.items() if k in update_fields}
    if not node:
        return False  # 或者 raise Exception("Entity not found")
    
    # 更新属性
    for k, v in entity_dict.items():
        node[k] = v
    
    # 保存更新
    graph.push(node)
    return True



def create_node_relation(entity,relationship):
    entity_dict = {}
    graph = get_graph()
    for k, v in entity.items():
    
        props = {prop: val for prop, val in {
            "entity_id": v.get("entity_id"),
            "entity_name": v.get("entity_name"),
            "entity_name_zh": v.get("entity_name_zh")
        }.items() if val not in (None, "")}

        if props.get("entity_id") is None:
            raise ValueError("entity_id is required")
        
        node = Node(v["entity_type"], **props)
        graph.merge(node, v["entity_type"], "entity_id")
        entity_dict[k] = node

    # query = """
    # MATCH (a:Disease {entity_id:$a_id}), (b:Gene {entity_id:$b_id})
    # MERGE (a)-[r:EVIDENCED_BY {association_id:$assoc_id}]->(b)
    # ON CREATE SET r += $props
    # ON MATCH SET  r += $props
    # """
    # graph.run(query, a_id="D0001", b_id="G0001", assoc_id="R1", props={"source":"PMID1"})
    association_id = relationship.get("association_id")
    predicate = relationship.get("predicate")
    # rel = Relationship(entity_dict["subject"],predicate, entity_dict["object"],**relationship,identity=association_id)

    # graph.create(rel)
    query = """
    MATCH (a {{entity_id:$a_id}}), (b {{entity_id:$b_id}})
    MERGE (a)-[r:{predicate} {{association_id:$assoc_id}}]->(b)
    ON CREATE SET r += $props
    ON MATCH SET r += $props
    RETURN r
    """.format(predicate=predicate)
    props = {k:v for k,v in relationship.items() if k != "association_id" and k != "predicate"}

    graph.run(query, a_id=entity_dict["subject"]["entity_id"],
                b_id=entity_dict["object"]["entity_id"],
                assoc_id=association_id,
                props=props)
    # relationship = {k: v for k, v in relationship.items() if k not in ["association_id"]}

    # # 赋值给关系对象
    # for k, v in relationship.items():
    #     rel[k] = v

    # # 再写入数据库
    # graph.push(rel)  # push 会更新已存在节点/关系的属性
    return entity_dict


from brave.api.schemas.project import AddProject,UpdateProject
from brave.api.models.core import t_project,samples as t_samples,analysis_result as t_analysis_result
import uuid
from fastapi import HTTPException

async def add_project(conn,AddProject:AddProject):
    uuid_str = str(uuid.uuid4())
    stmt = t_project.insert().values(project_id=uuid_str,project_name=AddProject.project_name,metadata_form=AddProject.metadata_form)
    conn.execute(stmt)
    return uuid_str

async def update_project(conn,UpdateProject:UpdateProject):
    stmt = t_project.update().where(t_project.c.project_id==UpdateProject.project_id).values(project_name=UpdateProject.project_name,metadata_form=UpdateProject.metadata_form)
    conn.execute(stmt)
    return UpdateProject.project_id

async def list_project(conn):
    stmt = t_project.select()
    result = conn.execute(stmt)
    return result.mappings().all()

async def find_by_project_id(conn,project_id:str):
    stmt = t_project.select().where(t_project.c.project_id==project_id)
    result = conn.execute(stmt).mappings().first()
    return result

def find_by_project_ids(conn,project_ids:list[str]):
    stmt = t_project.select().where(t_project.c.project_id.in_(project_ids))
    result = conn.execute(stmt).mappings().all()
    return result

async def delete_project(conn,project_id:str):
    stmt = t_samples.select().where(t_samples.c.project==project_id)
    result = conn.execute(stmt).fetchone()
    if result:
        raise HTTPException(status_code=400, detail="Project has samples, cannot be deleted")

    stmt = t_analysis_result.select().where(t_analysis_result.c.project==project_id)
    result = conn.execute(stmt).fetchone()
    if result:
        raise HTTPException(status_code=400, detail="Project has analysis results, cannot be deleted")


    stmt = t_analysis_result.select().where(t_analysis_result.c.project==project_id)
    result = conn.execute(stmt).fetchone()
    if result:
        raise HTTPException(status_code=400, detail="Project has analysis results, cannot be deleted")
    conn.execute(stmt)
    stmt = t_project.delete().where(t_project.c.project_id==project_id)
    conn.execute(stmt)
    return project_id

async def init_db(conn):
    stmt = t_project.select().where(t_project.c.project_id=="default")
    result = conn.execute(stmt).fetchone()
    if not result:
        project_dict = {"project_id":"default","project_name":"default","metadata_form":"[]"}
        stmt = t_project.insert().values(project_dict)
        conn.execute(stmt)



from fastapi import APIRouter
from brave.api.config.db import get_engine
from brave.microbe.service import study_service
from brave.microbe.utils import pubmed
study_api = APIRouter(prefix="/study")


@study_api.get("/mining-study/{entity_id}")
async def get_entity( entity_id: str):
    with get_engine().begin() as conn:
        result = study_service.mining_study(conn, entity_id)
    return result

@study_api.post("/get-fulltext/{entity_id}")
async def get_pmc_fulltext(entity_id: str):
    with get_engine().begin() as conn:
        study = study_service.mining_study(conn, entity_id)
        pmcid = study["pmcid"]
        fulltext = await pubmed.get_pmc_fulltext(pmcid)
        study_service.update_study(conn, entity_id, {"fulltext": fulltext})
    return {"message": "Full text updated successfully"}


# workflow_event_system/handlers/workflow_events.py

from dependency_injector.wiring import inject
from brave.api.core.routers.workflow_event_router import WorkflowEventRouter
from dependency_injector.wiring import inject, Provide
from brave.api.schemas.analysis import Analysis
from brave.api.service.result_parse.analysis_manage import AnalysisManage
from brave.app_container import AppContainer
from brave.api.core.event import WorkflowEvent
from brave.api.core.evenet_bus import EventBus
import brave.api.service.analysis_service as analysis_service
import asyncio
from brave.api.service.sse_service import SSESessionService
import json
@inject
def setup_handlers(
    evenet_bus:EventBus  = Provide[AppContainer.event_bus],
    router: WorkflowEventRouter = Provide[AppContainer.workflow_event_router],
    sse_service:SSESessionService = Provide[AppContainer.sse_service],
    result_parse_manage:AnalysisManage = Provide[AppContainer.result_parse_manage]
    ):


            
    @router.on_event(WorkflowEvent.ON_FLOW_BEGIN)
    async def on_flow_begin(analysis:Analysis,msg: dict):
        print(f"🚀 [on_flow_begin] {msg['analysis_id']}")
        await sse_service.push_message({"group": "default", "data": json.dumps(msg)})
        # result_parse_manage.create(analysis.analysis_id)

    @router.on_event(WorkflowEvent.ON_FILE_PUBLISH)
    async def on_file_publish(analysis:Analysis,msg: dict):
        print(f"✅ [on_file_publish] {msg['analysis_id']}")
        await sse_service.push_message({"group": "default", "data": json.dumps(msg)})
        # await result_parse_manage.parse(analysis.analysis_id)




    @router.on_event(WorkflowEvent.ON_PROCESS_COMPLETE)
    async def on_process_complete(analysis:Analysis,msg: dict):
        print(f"🚀 [on_process_complete] {msg['analysis_id']}")
        await sse_service.push_message({"group": "default", "data": json.dumps(msg)})
        # await result_parse_manage.parse(analysis.analysis_id)


    @router.on_event(WorkflowEvent.ON_FLOW_COMPLETE)
    async def on_flow_complete(analysis:Analysis,msg: dict):
        print(f"🚀 [on_flow_complete] {msg['analysis_id']}")
        await sse_service.push_message({"group": "default", "data": json.dumps(msg)})
        # await result_parse_manage.parse(analysis.analysis_id)

        # result_parse_manage.remove(analysis.analysis_id)
        


    # @router.on_event(WorkflowEvent.ON_JOB_SUBMITTED)
    # async def on_job_submitted(analysis:Analysis,msg: dict):
    #     print(f"🚀 [on_job_submitted] {msg['job_id']}")






import json
import os
import uuid
from brave.api.models.core import t_namespace
from brave.api.schemas.namespace import SaveNamespace
import  brave.api.service.namespace_service as namespace_service
from fastapi import APIRouter, HTTPException
from brave.api.config.db import get_engine
import brave.api.service.pipeline as pipeline_service
import brave.api.service.container_service as container_service

namespace = APIRouter()

@namespace.post("/save-or-update-namespace",tags=['namespace'])
async def save_namespace_controller(saveNamespace:SaveNamespace):
    if saveNamespace.volumes:
        try:
            json.loads(saveNamespace.volumes)
        except Exception as e:
            raise RuntimeError(f"The volumes field is not in legal JSON format: {e}")
    with get_engine().begin() as conn:
        if saveNamespace.namespace_id:
            find_namespace = namespace_service.find_namespace(conn,saveNamespace.namespace_id)

            namespace_id = find_namespace.namespace_id 
            namespace_service.update_namespace(conn,saveNamespace.namespace_id,saveNamespace.dict())
        else:
            str_uuid = str(uuid.uuid4())
            saveNamespace.namespace_id = str_uuid
            namespace_id = str_uuid
            namespace_service.save_namespace(conn,saveNamespace.model_dump())
    
    # namespace = f"{pipeline_dir}/{namespace_id}"
    # if not os.path.exists(namespace):
    #     os.makedirs(namespace)
    # with open(f"{namespace}/namespace.json","w") as f:
    #     f.write(json.dumps(saveNamespace.dict()))
    # namespace_service.write_namespace(namespace_id,saveNamespace.model_dump())
    return {"message":"success"}

@namespace.get("/list-namespace",tags=['namespace'])
async def list_namespace():
    with get_engine().begin() as conn:
        return namespace_service.list_namespace(conn)

    

@namespace.delete("/delete-namespace-by-id/{namespace_id}",tags=['namespace'])
async def delete_namespace_by_namespace_id(namespace_id:str):
    with get_engine().begin() as conn:
        find_namespace = namespace_service.find_namespace(conn,namespace_id)
      
        if find_namespace:
            namespace_service.delete_namespace(conn,namespace_id)
            return {"message":"success"}


@namespace.get("/find-namespace-by-id/{namespace_id}",tags=['namespace'])
async  def find_by_id(namespace_id):
    with get_engine().begin() as conn:
        find_namespace = namespace_service.find_namespace(conn,namespace_id)
        if not find_namespace:
            raise HTTPException(status_code=404, detail=f"Namespace with id {namespace_id} not found")
        return find_namespace


@namespace.get("/get-used-namespace",tags=['namespace'])
async def get_used_namespace():
    with get_engine().begin() as conn:
        find_namespace = namespace_service.get_used_namespace(conn)
        return find_namespace or {}
        # if not find_namespace:
        #     raise HTTPException(status_code=404, detail=f"Namespace with id default not found")
        # return find_namespace
@namespace.post("/set-used-namespace/{namespace_id}",tags=['namespace'])
async def set_used_namespace(namespace_id:str):
    with get_engine().begin() as conn:
        find_namespace = namespace_service.find_namespace(conn,namespace_id)
        if not find_namespace:
            raise HTTPException(status_code=404, detail=f"Namespace with id {namespace_id} not found")
        return namespace_service.set_used_namespace(conn,namespace_id)


import asyncio
from dataclasses import dataclass, field
from typing import Dict, Any
import time

from brave.api.core.pubsub import PubSubManager
from brave.api.core.routers.workflow_event_router import WorkflowEventRouter
from brave.api.core.event import WorkflowEvent

@dataclass
class WorkflowQueue:
    queue: asyncio.Queue
    task: asyncio.Task
    last_active: float = field(default_factory=time.time)
    # subscribers: int = 0

class WorkflowQueueManager:
    def __init__(self, pubsub: PubSubManager, workflow_event_router: WorkflowEventRouter):
        self.workflow_map: dict[str, WorkflowQueue] = {}
        self.pubsub = pubsub
        self.workflow_event_router = workflow_event_router
    

    def register_subscriber(self, topic: str, callback):
        self.pubsub.subscribe(topic, callback)

    def register(self, workflow_id: str):
        if workflow_id not in self.workflow_map:
            queue = asyncio.Queue()
            task = asyncio.create_task(self._consume_loop(workflow_id, queue))
            self.workflow_map[workflow_id] = WorkflowQueue(queue=queue, task=task)
    
 
        

    async def put(self, workflow_id: str, msg: dict):
        self.register(workflow_id)
        wfq = self.workflow_map[workflow_id]
        wfq.last_active = time.time()
        await wfq.queue.put(msg)

    async def dispatch(self, msg: dict):
        workflow_id = msg.get("analysis_id", "global")
        await self.put(workflow_id, msg)

    def get(self, workflow_id: str) -> WorkflowQueue:
        return self.workflow_map[workflow_id]

    async def _consume_loop(self, workflow_id: str, queue: asyncio.Queue):
        while True:
            msg = await queue.get()
            try:
                await self.pubsub.publish(workflow_id, msg)
                
                try:
                    event = WorkflowEvent(msg.get("workflow_event"))
                except ValueError:
                    event = msg.get("workflow_event")
                    print(f"[WorkflowEventRouter] Unknown event type '{event}'", msg)
                    return

                await self.workflow_event_router.dispatch(event,workflow_id,msg)
            except Exception as e:
                print(f"[Consumer ERROR] workflow {workflow_id}: {e}")

    async def cleanup(self, timeout: int = 300):
        now = time.time()
        to_delete = []
        for wf_id, wfq in self.workflow_map.items():
            if  wfq.queue.empty() and (now - wfq.last_active > timeout):
                wfq.task.cancel()
                to_delete.append(wf_id)
            
        for wf_id in to_delete:
            del self.workflow_map[wf_id]
            
            await self.pubsub.publish(wf_id, {"event": "workflow_cleanup", "workflow_id": wf_id})
            # await self.workflow_event_router.dispatch(WorkflowEvent.WORKFLOW_CLEANUP, {"analysis_id": wf_id})
            print(f"[Cleanup] Removed workflow {wf_id}")

    
    async def cleanup_loop(self):
        while True:
            await self.cleanup()
            await asyncio.sleep(10)



from dependency_injector.wiring import Provide, inject
from fastapi import APIRouter, Depends,HTTPException
from importlib.resources import files, as_file

import json
import os
import glob
from brave.api.config.config import get_settings
from brave.api.enum.component_script import ScriptName
from brave.api.service.pipeline import get_pipeline_dir,get_pipeline_list
from collections import defaultdict
from brave.api.models.core import t_pipeline_components,t_pipeline_components_relation
import uuid
from brave.api.config.db import get_engine
from sqlalchemy import or_, select, and_, join, func,insert,update
import re
from brave.api.schemas.pipeline import InstallComponent, PagePipelineQuery, SavePipeline,Pipeline,QueryPipeline,QueryModule, SavePipelineComponentsEdges,SavePipelineRelation,SaveOrder
import brave.api.service.pipeline  as pipeline_service
from sqlalchemy import  Column, Integer, String, Text, select, cast, null,text,case
from sqlalchemy.orm import aliased
from sqlalchemy.sql import union_all
from brave.api.service.sse_service import SSESessionService
import brave.api.utils.service_utils  as service_utils
import asyncio
import time
from starlette.concurrency import run_in_threadpool
from typing import List
import brave.api.service.container_service as container_service
from brave.app_container import AppContainer
import brave.api.service.notebook as notebook_service
import shutil
from fastapi import  File, UploadFile
import shutil
import time
from brave.api.executor.base import JobExecutor


pipeline = APIRouter()

def camel_to_snake(name):
    s1 = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


# pipeline,software,file,downstream
# BASE_DIR = os.path.dirname(__file__)
@pipeline.post("/import-pipeline",tags=['pipeline'])
async def import_pipeline():
    pipeline_files = get_pipeline_list()
    new_pipeline_components_list = [] 
    new_pipeline_components_relation_list = []
    with get_engine().begin() as conn:
        pipeline_list = find_db_pipeline(conn, "pipeline")
        db_pipeline_key_list = [item.install_key for item in pipeline_list ]
        for pipeline_item_file in pipeline_files:
            json_data = get_pipeine_content(pipeline_item_file)
            install_key = os.path.basename( os.path.dirname(pipeline_item_file))
            if install_key in db_pipeline_key_list:
                continue
            pipeline_item = {k:v for k,v in json_data.items() if  k !="items"}
            pipeline_components_id =install_key# str(uuid.uuid4())
            # pipeline_id  = pipeline_components_id
            new_pipeline_components_list.append({
                "component_id":pipeline_components_id,
                "install_key":install_key,
                # "pipeline_key":wrap_pipeline_key,
                # "parent_pipeline_id":"0",
                "component_type":"pipeline",
                "content":json.dumps(pipeline_item)
            })

            keys_to_remove = [ 'inputFile','outputFile']
            for analysis_software in json_data['items']:
                analysis_software_ = {k:v for k,v in analysis_software.items() if  k not in keys_to_remove}
                analysis_software_uuid = str(uuid.uuid4())
                new_pipeline_components_list.append({
                    "component_id":analysis_software_uuid,
                    "install_key":install_key,
                    # "parent_pipeline_id":wrap_pipeline_uuid,
                    # "pipeline_key":wrap_pipeline_key,
                    "component_type":"software",
                    "content":json.dumps(analysis_software_)
                })
                new_pipeline_components_relation_list.append({
                    "relation_type":"pipeline_software",
                    "install_key":install_key,
                    "component_id":analysis_software_uuid,
                    "parent_component_id":pipeline_components_id,
                    # "pipeline_id":pipeline_id
                })
                for key in keys_to_remove:
                    add_analysis_file( analysis_software_uuid,install_key,analysis_software,key,new_pipeline_components_list,new_pipeline_components_relation_list)
                # key= "parseAnalysisResultModule"
        insert_stmt = insert(t_pipeline_components).values(new_pipeline_components_list)
        conn.execute(insert_stmt)
        insert_stmt = insert(t_pipeline_components_relation).values(new_pipeline_components_relation_list)
        conn.execute(insert_stmt)
        return {
            "pipeline":new_pipeline_components_list,
            "relation_pipeline":new_pipeline_components_relation_list
        }
        

def add_analysis_file(analysis_software_uuid,install_key,pipeline_item,key,new_pipeline_components_list,new_pipeline_components_relation_list):
    if key in pipeline_item:
              
        for analysis_file in pipeline_item[key]:
            analysis_file_uuid = str(uuid.uuid4())  
            # if key=='downstreamAnalysis':
            analysis_file_ = {k:v for k,v in analysis_file.items() if  k !="downstreamAnalysis"}
            new_pipeline_components_list.append({
                "component_id":analysis_file_uuid,
                "install_key":install_key,
                # "parent_pipeline_id":pipeline_uuid,
                # "pipeline_key":wrap_pipeline_key,
                "component_type":"file",
                "content":json.dumps(analysis_file_)
            })
            if key=="inputFile":
                new_pipeline_components_relation_list.append({
                    "relation_type":"software_input_file",
                    "install_key":install_key,
                    "component_id":analysis_file_uuid,
                    # "pipeline_id":pipeline_id,
                    "parent_component_id":analysis_software_uuid
                }) 
            elif  key=="outputFile":
                new_pipeline_components_relation_list.append({
                    "relation_type":"software_output_file",
                    "install_key":install_key,
                    "component_id":analysis_file_uuid,
                    # "pipeline_id":pipeline_id,
                    "parent_component_id":analysis_software_uuid
                })
            if "downstreamAnalysis" in analysis_file:
                for downstream_analysis in analysis_file["downstreamAnalysis"]:
                    downstream_analysis_uuid = str(uuid.uuid4()) 
                    new_pipeline_components_list.append({
                        "component_id":downstream_analysis_uuid,
                        "install_key":install_key,
                        # "parent_pipeline_id":item_uuid,
                        # "pipeline_key":wrap_pipeline_key,
                        "component_type":"downstream",
                        "content":json.dumps(downstream_analysis)
                    })
                    new_pipeline_components_relation_list.append({
                        "relation_type":"file_script",
                        # "pipeline_id":pipeline_id,
                        "install_key":install_key,
                        "component_id":downstream_analysis_uuid,
                        "parent_component_id":analysis_file_uuid
                    })

            # else:
            # new_pipeline_list.append({
            #     "pipeline_id":item_uuid,
            #     "parent_pipeline_id":pipeline_uuid,
            #     "pipeline_key":wrap_pipeline_key,
            #     "pipeline_type":camel_to_snake(key),
            #     "content":json.dumps(item)
            # })

def find_db_pipeline(conn, component_type):
    return conn.execute(t_pipeline_components.select() 
        .where(t_pipeline_components.c.component_type==component_type)).fetchall()

@pipeline.get("/get-pipeline/{name}",tags=['pipeline'])
async def get_pipeline(name):
    pipeline_dir =  get_pipeline_dir()
    

    # filename = f"{name}.json"
    json_file = f"{pipeline_dir}/{name}/main.json"
    data = {
        # "files":json_file,
        # # "wrapAnalysisPipeline":name,
        # "exists":os.path.exists(json_file)
    }
    if os.path.exists(json_file):
        json_data = get_pipeine_content(json_file)
        data.update(json_data)
    return data

def get_pipeline_item(item):
    content= json.loads(item.content)
    return {
        "id":item.id,
        "pipeline_id":item.pipeline_id,
        "pipeline_key":item.pipeline_key,
        "parent_pipeline_id":item.parent_pipeline_id,
        "pipeline_order":item.pipeline_order,
        "pipeline_type":item.pipeline_type,
        **content
    }
@pipeline.get("/get-component-parent/{component_id}",tags=['pipeline'])
async def get_component_parent(component_id,component_type):
    

    base = select(
        t_pipeline_components.c.component_id,
        t_pipeline_components.c.component_type,
        t_pipeline_components.c.install_key,
        t_pipeline_components.c.content,
        t_pipeline_components.c.component_name,
        t_pipeline_components.c.tags,
        cast(null(), String(255)).label("relation_type"),
        cast(null(), String(255)).label("parent_component_id"),
        cast(null(), Integer).label("order_index"),
        cast(null(), String(255)).label("relation_id"),
    ).where(
        t_pipeline_components.c.component_id == component_id,
        t_pipeline_components.c.component_type == component_type,
    )

    tp1 = aliased(t_pipeline_components)
    rel = t_pipeline_components_relation
    # fp = aliased(cte)  # 引用递归CTE自身
    # base_alias = base.alias()

    stmt_parenet = select(
        tp1.c.component_id,
        tp1.c.component_type,
        tp1.c.install_key,
        tp1.c.content,
        tp1.c.component_name,
        tp1.c.tags,
        rel.c.relation_type,
        rel.c.parent_component_id,
        rel.c.order_index,
        rel.c.relation_id,
    ).select_from(
        tp1.join(rel, tp1.c.component_id == rel.c.parent_component_id)
    ).where(
        rel.c.component_id == component_id
    )

    # # 合并 base 和 recursive，生成完整的递归CTE
    stmt = base.union_all(stmt_parenet) 
    with get_engine().begin() as conn:
        data = conn.execute(stmt).mappings().all()
    child_item  = next((item for item in data if item["component_type"] == component_type),None)
    if not child_item:
        raise HTTPException(status_code=500, detail=f"{component_id}没有找到!")  

    if component_type == "script":
        child_item = {
            **child_item,
            **json.loads(child_item["content"])
        }
        del child_item["content"]

    parent_item_list = [dict(item) for item in data if item['component_type'] != component_type]
    # resul_dict= {}
    # resul_dict['script'] = dict(child_item)
    child_item['parent'] = parent_item_list
    return child_item


@pipeline.get("/get-pipeline-dag/{pipeline_id}",tags=['pipeline'])
async def get_pipeline_dag(pipeline_id):
    s_alias = t_pipeline_components.alias("s")
    t_alias = t_pipeline_components.alias("t")
    stmt1 = (
        select(
            t_pipeline_components_relation.c.parent_component_id.label("source"),
            t_pipeline_components_relation.c.component_id.label("target"),
            s_alias.c.component_name.label("source_name"),
            t_alias.c.component_name.label("target_name"),
        )
        .select_from(
            t_pipeline_components_relation
            .join(s_alias, t_pipeline_components_relation.c.parent_component_id == s_alias.c.component_id)
            .join(t_alias, t_pipeline_components_relation.c.component_id == t_alias.c.component_id)
        )
        .where(
            and_(
                t_pipeline_components_relation.c.relation_type == "pipeline_software",
                t_pipeline_components_relation.c.pipeline_id == pipeline_id
            )
        )
    )
    stmt2 = (
        select(t_pipeline_components)
        .distinct()
        .select_from(
            t_pipeline_components_relation
            .join(
                t_pipeline_components,
                or_(
                    t_pipeline_components.c.component_id == t_pipeline_components_relation.c.parent_component_id,
                    t_pipeline_components.c.component_id == t_pipeline_components_relation.c.component_id
                )
            )
        )
        .where(
            and_(
                t_pipeline_components_relation.c.relation_type == "pipeline_software",
                t_pipeline_components_relation.c.pipeline_id == pipeline_id
            )
        )
    )
    stmt3 = select(t_pipeline_components).where(t_pipeline_components.c.component_id == pipeline_id)
    with get_engine().begin() as conn:
        data = conn.execute(stmt1).mappings().all()
        data2 = conn.execute(stmt2).mappings().all()
        data3 = conn.execute(stmt3).mappings().first()
    return {
        **data3,
        "edges":data,
        "nodes":data2
    }

@pipeline.get("/get-pipeline-v2/{name}",tags=['pipeline'])
async def get_pipeline_v2(name,component_type="pipeline"):
    with get_engine().begin() as conn:
        return pipeline_service.get_pipeline_v2(conn,name,component_type)


  
@pipeline.get("/get-component-module-content/{component_id}",tags=['pipeline'])
async def get_module_content(component_id,script_name:ScriptName):
    # module_dir = queryModule.component_id
    with get_engine().begin() as conn:
        find_component = pipeline_service.find_pipeline_by_id(conn,component_id)
        find_component = {
            **{k:v for k,v in find_component.items() if k != "content"},
            **json.loads(find_component["content"])
        }
        if not find_component:
            raise HTTPException(status_code=500, detail=f"根据{component_id}不能找到记录!")
    # if queryModule.module_dir:
    #     module_dir = queryModule.module_dir
    module_info:dict = pipeline_service.find_component_module(find_component,script_name)
    # py_module_path = py_module['path']
    if module_info and os.path.exists(module_info['path']):
        with open(module_info['path'],"r") as f:
            module_content = f.read()
    # py_module['content'] = py_module_content
    return {
        "path":module_info['path'], 
        "content":module_content
    }
    

def get_pipeine_content(json_file):
    markdown_dict = get_all_markdown()
    with open(json_file,"r") as f:
        json_data = json.load(f)
        # update_downstream_markdown(json_data.items)
        for item1 in  json_data['items']:
            if "markdown" in item1:
                content = get_markdown_content(markdown_dict,item1['markdown'] )
                item1['markdown'] = content
            if "downstreamAnalysis" in item1:
                for item2 in item1['downstreamAnalysis']:
                    if "markdown" in item2:
                        content = get_markdown_content(markdown_dict,item2['markdown'] )
                        item2['markdown'] = content
    return json_data

def get_config():
    pipeline_dir =  get_pipeline_dir()
    config = f"{pipeline_dir}/config.json"
    if os.path.exists(config):
        with open(config,"r") as f:
            return json.load(f)
    else:
        return {}
    

def get_category(name,key):
    config = get_config()
    if "category" in config:
        category = config['category']
        if name in category:
            return category[name][key]
    return name

def get_pipeline_one_v2(item):
    try:
        data = json.loads(item.content)
        result = {
            "id":item.id,
            "component_id":item.component_id,
            "path":item.component_id,
            "name":data['name'],
            "category":data['category'],
            "img":f"/brave-api/img/{data['img']}",
            "tags":data['tags'],
            "description":data['description'] if 'description' in data else "",
            "order":item.order_index
        }
        return  result
    except (ValueError, TypeError):
        return {
            "id":item.id,
            "pipeline_id":item.component_id,
            "path":item.component_id,
            "name":"unkonw",
            "category":"unkonw",
            "img":f"/brave-api/img/unkonw",
            "tags":["unkonw"],
            "description":"unkonw",
            "order":item.order_index
        }       
@pipeline.get("/list-pipeline-v2",tags=['pipeline'])
async def list_pipeline_v2():
    with get_engine().begin() as conn:
        wrap_pipeline_list = find_db_pipeline(conn, "pipeline")
        pipeline_list = [get_pipeline_one_v2(item) for item in wrap_pipeline_list]
        # pipeline_list = sorted(pipeline_list, key=lambda x:x["order"] if x["order"] is not None else x["id"])
    
    grouped = defaultdict(list)
    for item in pipeline_list:
        grouped[item["category"]].append(item)

    result = []
    for category, items in grouped.items():
        result.append({
            "name": get_category(category,"name"),
            "items": items
        })
    return result
    # pass

def get_pipeline_one(item):
    with open(item,"r") as f:
        data = json.load(f)
    data = {
        "path":os.path.basename(os.path.dirname(item)),
        "name":data['name'],
        "category":data['category'],
        "img":f"/brave-api/img/{data['img']}",
        "tags":data['tags'],
        "description":data['description'],
        "order":data['order']
    }
    return data

# @pipeline.get("/list-pipeline",tags=['pipeline'])
# async def get_pipeline():
#     # json_file = str(files("brave.pipeline.config").joinpath("config.json"))
#     # with open(json_file,"r") as f:
#     #     config = json.load(f)
#     # pipeline_files = files("brave.pipeline")
#     pipeline_files = get_pipeline_list()
#     pipeline_files = [get_pipeline_one(str(item)) for item in pipeline_files]
#     pipeline_files = sorted(pipeline_files, key=lambda x: x["order"])
#     grouped = defaultdict(list)
#     for item in pipeline_files:
#         grouped[item["category"]].append(item)

#     result = []
#     for category, items in grouped.items():
#         result.append({
#             "name": get_category(category,"name"),
#             "items": items
#         })
#     return result


def get_pipeline_file(filename):
    nextflow_dict = get_all_pipeline()
    if filename not in nextflow_dict:
        raise HTTPException(status_code=500, detail=f"{filename}不存在!")  
    return nextflow_dict[filename]

def get_all_pipeline():
    pipeline_dir =  get_pipeline_dir()
    nextflow_list = glob.glob(f"{pipeline_dir}/*/nextflow/*.nf")
    nextflow_dict = {os.path.basename(item).replace(".nf",""):item for item in nextflow_list}
    return nextflow_dict



def get_all_markdown():
    pipeline_dir =  get_pipeline_dir()
    markdown_list = glob.glob(f"{pipeline_dir}/*/markdown/*.md")
    markdown_dict = {os.path.basename(item).replace(".md",""):item for item in markdown_list}
    return markdown_dict

def get_markdown_content(markdown_dict,name):
    markdown_file = markdown_dict[name]
    with open(markdown_file,"r") as f:
        content = f.read()
    return content

def get_downstream_analysis(item):
    with open(item,"r") as f:
        data = json.load(f)
    file_list = [
        item
        for d in data['items']
        if "downstreamAnalysis" in d
        for item in d['downstreamAnalysis']
    ]

    return file_list

@pipeline.get("/find_downstream_analysis/{analysis_method}",tags=['pipeline'])
async def get_downstream_analysis_list(analysis_method):
    pipeline_files = get_pipeline_list()
    downstream_list = [get_downstream_analysis(item) for item in pipeline_files]
    downstream_list = [item for sublist in downstream_list for item in sublist]
    downstream_dict = {item['saveAnalysisMethod']: item for item in downstream_list  if 'saveAnalysisMethod' in item}
    return downstream_dict[analysis_method]
    
@pipeline.post("/find-pipeline",tags=['pipeline'])
async def find_pipeline_by_id(queryPipeline:QueryPipeline):
    with get_engine().begin() as conn:
        return pipeline_service.find_pipeline_by_id(conn,queryPipeline.component_id)

@pipeline.post("/list-pipeline-components",tags=['pipeline'],response_model=list[Pipeline])
async def list_pipeline(queryPipeline:QueryPipeline):
    with get_engine().begin() as conn:
        return pipeline_service.list_pipeline(conn,queryPipeline)

@pipeline.post("/page-pipeline-components",tags=['pipeline'])
async def page_pipeline(query:PagePipelineQuery ):
    with get_engine().begin() as conn:
        return pipeline_service.page_pipeline(conn,query)


@pipeline.post("/save-pipeline-components-edges",tags=['pipeline'])
async def save_pipeline_components_edges(savePipelineComponentsEdges:SavePipelineComponentsEdges):
    with get_engine().begin() as conn:
        pipeline_service.save_pipeline_components_edges(conn,savePipelineComponentsEdges)
    return {"message":"success"}    

# def get_pipeline_id_by_parent_id(conn, start_id: str) -> str | None:
#     sql = text("""
#         WITH RECURSIVE ancestor_path AS (
#             SELECT
#                 pipeline_id,
#                 parent_pipeline_id,
#                 relation_type
#             FROM relation_pipeline
#             WHERE pipeline_id = :start_id

#             UNION ALL

#             SELECT
#                 rp.pipeline_id,
#                 rp.parent_pipeline_id,
#                 rp.relation_type
#             FROM relation_pipeline rp
#             JOIN ancestor_path ap ON rp.pipeline_id = ap.parent_pipeline_id
#         )
#         SELECT pipeline_id
#         FROM ancestor_path
#         WHERE relation_type = 'pipeline_software'
#         LIMIT 1;
#     """)

#     result = conn.execute(sql, {"start_id": start_id})
#     row = result.first()
#     return row[0] if row else None

@pipeline.post("/find-pipeline-relation/{relation_id}",tags=['pipeline'])
async def find_pipeline_relation(relation_id):
    with get_engine().begin() as conn:    
        stmt = t_pipeline_components_relation.select().where(t_pipeline_components_relation.c.relation_id == relation_id)
        return conn.execute(stmt).mappings().first()


@pipeline.post("/save-pipeline-relation",tags=['pipeline'])
async def save_pipeline_relation_controller(savePipelineRelation:SavePipelineRelation):
    with get_engine().begin() as conn:  
        await save_pipeline_relation(conn,savePipelineRelation)

        
        return {"message":"success"}

async def save_pipeline_relation(conn,savePipelineRelation):
    save_pipeline_relation_dict = savePipelineRelation.dict()
    # save_pipeline_relation_dict = {k:v for k,v in save_pipeline_relation_dict.items() if k!="pipeline_id"}
    if savePipelineRelation.parent_component_id:
        parent_component = pipeline_service.find_pipeline_by_id(conn,savePipelineRelation.parent_component_id)
        if parent_component:
            component_id = parent_component["component_id"]
    if savePipelineRelation.relation_id:
        stmt = t_pipeline_components_relation.update().values(save_pipeline_relation_dict).where(t_pipeline_components_relation.c.relation_id==savePipelineRelation.relation_id)
    else:
        save_pipeline_relation_dict['relation_id'] = str(uuid.uuid4())
        child_component_count = pipeline_service.get_child_component_count(conn,savePipelineRelation.parent_component_id,savePipelineRelation.relation_type)
        save_pipeline_relation_dict['order_index'] = child_component_count + 1
        stmt = t_pipeline_components_relation.insert().values(save_pipeline_relation_dict)
        conn.execute(stmt)
    
    pipeline_service.write_component_json(component_id)

    # stmt = t_pipeline_components.select().where(t_pipeline_components.c.component_id ==savePipelineRelation.component_id)
    # find_pipeine = conn.execute(stmt).fetchone()
    # await run_in_threadpool(create_pipeline_dir, savePipelineRelation.pipeline_id, find_pipeine.content ,find_pipeine.component_type)
    # pipeline_service.create_wrap_pipeline_dir(savePipelineRelation.pipeline_id)
    # pipeline_service.create_file(savePipelineRelation.pipeline_id, find_pipeine.component_type,content)
    # content = json.loads(find_pipeine.content)



@pipeline.post("/save-pipeline",tags=['pipeline'])
async def save_pipeline(savePipeline:SavePipeline):
    try:
        json.loads(savePipeline.content)
    except Exception as e:
        # print("component content json error",json.dumps(savePipeline,indent=4))
        # raise HTTPException(status_code=500, detail=f"组件内容不是合法的json格式! 错误信息:{str(e)}")
        raise ValueError(f"The component content is not valid JSON format! Error message:{str(e)}")
 
 
    save_pipeline_dict = savePipeline.dict()
    save_pipeline_dict = {k:v for k,v in save_pipeline_dict.items() if k!="parent_component_id" and k!="pipeline_id" and k!='relation_type' }
    
    with get_engine().begin() as conn:
        find_pipeine = None
        if savePipeline.component_id:
            stmt = t_pipeline_components.select().where(t_pipeline_components.c.component_id == savePipeline.component_id)
            find_pipeine = conn.execute(stmt).fetchone()

            if not find_pipeine:
                raise HTTPException(status_code=500, detail=f"根据{savePipeline.component_id}不能找到记录!")
            component_id = find_pipeine.component_id

        if find_pipeine:
            save_pipeline_dict = {k:v for k,v in save_pipeline_dict.items() if k!="component_id" and v is not  None } 
            stmt = t_pipeline_components.update().values(save_pipeline_dict).where(t_pipeline_components.c.component_id==savePipeline.component_id)
            conn.execute(stmt)
            
        else:
           
                
            str_uuid = str(uuid.uuid4())  
            save_pipeline_dict['component_id'] = str_uuid
            component_id = str_uuid
            stmt = t_pipeline_components.insert().values(save_pipeline_dict)
            conn.execute(stmt)
            
            if savePipeline.relation_type:
                await save_pipeline_relation(conn, SavePipelineRelation(
                    component_id=component_id,
                    parent_component_id= savePipeline.parent_component_id,
                    relation_type=savePipeline.relation_type,
                    # pipeline_id=savePipeline.pipeline_id
                ))
     
    pipeline_service.write_component_json(component_id)
    # t0 = time.time()
    

    # await asyncio.sleep(0.5)
    # print("文件创建耗时", time.time() - t0)

    return {"message":"success"}



@pipeline.delete("/delete-pipeline-relation/{relation_id}")
async def delete_pipeline_relation(relation_id: str):
    component_id = None
    with get_engine().begin() as conn:
        component_relation = pipeline_service.find_by_relation_id(conn,relation_id)
        if not component_relation:
            raise HTTPException(status_code=500, detail=f"根据{relation_id}不能找到记录!") 
        parent_component = pipeline_service.find_pipeline_by_id(conn,component_relation.parent_component_id)
        if parent_component:
            component_id = parent_component["component_id"]
        stmt = t_pipeline_components_relation.delete().where(t_pipeline_components_relation.c.relation_id == relation_id)
        conn.execute(stmt)
    if component_id:
        pipeline_service.write_component_json(component_id)
    return {"message":"success"}



@pipeline.delete("/delete-component/{component_id}")
async def delete_component(component_id: str):

    with get_engine().begin() as conn:
        find_component = pipeline_service.find_component_by_id(conn,component_id)
        if not find_component:
            raise HTTPException(status_code=500, detail=f"Cannot find component for {component_id}!")
        stmt = t_pipeline_components_relation.select().where(t_pipeline_components_relation.c.parent_component_id ==component_id)
        parent_find_pipeine = conn.execute(stmt).fetchone()
        stmt = t_pipeline_components_relation.select().where(t_pipeline_components_relation.c.component_id ==component_id)
        child_find_pipeine = conn.execute(stmt).fetchall()

        if  parent_find_pipeine or child_find_pipeine:
            raise HTTPException(status_code=500, detail=f"Cannot delete because there are existing associations!") 
        else:
            # find_component = pipeline_service.find_component_by_id(conn,component_id)
            stmt = t_pipeline_components.delete().where(t_pipeline_components.c.component_id == component_id)
            conn.execute(stmt)
            # pipeline_service.delete_wrap_pipeline_dir(component_id)

    pipeline_service.delete_component_file(find_component)

    return {"message":"success"}

@pipeline.get("/find-by-component-id/{component_id}",tags=['pipeline'])
async def find_by_components_id(component_id):
    with get_engine().begin() as conn:
        stmt = t_pipeline_components.select().where(t_pipeline_components.c.component_id == component_id)
        return conn.execute(stmt).mappings().first()

    

@pipeline.post("/install-components",tags=['pipeline'])
@inject
async def install_component(installComponent:InstallComponent,force:bool=False,
    job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector])
):
    pipeline_dir = pipeline_service.get_pipeline_dir()
    with open(installComponent.path,"r") as f:
        data = json.load(f)
    path = os.path.dirname(installComponent.path)

    
    with get_engine().begin() as conn:
        pipeline_service.import_component(conn,path,force)
        pipeline_service.import_component_relation(conn,path,force)
        container_service.import_container(conn,path,force)
    
    install_path = f"{pipeline_dir}/{data['component_type']}/{data['component_id']}"

    if not os.path.exists(install_path):
        # os.makedirs(install_path)
        shutil.copytree(path, install_path)
        print("copytree",path, install_path)
    else:
        if force:   
            shutil.rmtree(install_path)
            shutil.copytree(path, install_path)
            print("force copytree",path, install_path)
    

    
    asyncio.create_task(job_executor.update_images_status())
    return {"message":"success"}

    





@pipeline.get("/get-depend-component/{component_id}",tags=['pipeline'])
async def get_depend_component(component_id):
    with get_engine().begin() as conn:
        find_component = pipeline_service.find_pipeline_by_id(conn, component_id)
        if not find_component:
            raise HTTPException(status_code=404, detail=f"Component {component_id} not found")
        child_depend_component = pipeline_service.get_child_depend_component(conn, component_id)
        parent_depend_component = pipeline_service.get_parent_depend_component(conn, component_id)
        return list(child_depend_component) + list(parent_depend_component)


@pipeline.post("/save-component-relation-order",tags=['pipeline'])
async def save_component_relation_order(saveOrder:list[SaveOrder]):
    with get_engine().begin() as conn:
        pipeline_service.save_order(conn,saveOrder)
    return {"message":"success"}


@pipeline.post("/update-component-description/{component_id}",tags=['pipeline'])
async def update_component_description(component_id,description ):
    with get_engine().begin() as conn:
        pipeline_service.update_component_description(conn,component_id,description)
    return {"message":"success"}





@pipeline.post("/component/convert-ipynb/{component_id}",tags=['pipeline'])
async def convert_ipynb(component_id):
    # module_dir = queryModule.component_id
    with get_engine().begin() as conn:
        find_component = pipeline_service.find_pipeline_by_id(conn,component_id)
        find_component = {
            **{k:v for k,v in find_component.items() if k != "content"},
            **json.loads(find_component["content"])
        }
        if not find_component:
            raise HTTPException(status_code=500, detail=f"根据{component_id}不能找到记录!")
    module_info:dict = pipeline_service.find_component_module(find_component, ScriptName.main)
    script_path  = module_info['path']
    ipynb_path = os.path.dirname(script_path)
    ipynb_path = f"{ipynb_path}/main.ipynb"
    if os.path.exists(script_path) and os.path.exists(ipynb_path):
        shutil.copy(script_path, f"{script_path}.tmp")
        notebook_service.convert_notebook(ipynb_path, script_path)
        return "success"
    raise HTTPException(status_code=404, detail=f"{script_path}或{ipynb_path}不存在!")


# @analysis_api.post("/analysis/convert-ipynb/{analysis_id}")
# async def convert_ipynb(analysis_id):
#     with get_engine().begin() as conn:
#         find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
#     script_path  = find_analysis["pipeline_script"]
#     ipynb_path = os.path.dirname(script_path)
#     ipynb_path = f"{ipynb_path}/main.ipynb"
#     if os.path.exists(script_path) and os.path.exists(ipynb_path):
#         shutil.copy(script_path, f"{script_path}.tmp")
#         notebook_service.convert_notebook(ipynb_path, script_path)
#         return "success"
#     raise HTTPException(status_code=404, detail=f"{script_path}或{ipynb_path}不存在!")


@pipeline.post("/component/upload/{component_id}")
async def upload_image(component_id,file: UploadFile = File(...)):
    # 限制只能上传图片类型
    if not file.content_type.startswith("image/"):
        raise HTTPException(status_code=404, detail=f"只允许上传图片文件!")
    with get_engine().begin() as conn:
        find_component = pipeline_service.find_pipeline_by_id(conn,component_id)
        find_component = {
            **{k:v for k,v in find_component.items() if k != "content"},
            **json.loads(find_component["content"])
        }
        if not find_component:
            raise HTTPException(status_code=500, detail=f"根据{component_id}不能找到记录!")
    module_info:dict = pipeline_service.find_component_module(find_component, ScriptName.main)
    script_path  = module_info['path']
    file_path = os.path.dirname(script_path)
    name, ext = os.path.splitext(file.filename)

    # file_path = os.path.join(file_path, file.filename)
    filename = f"main{ext}"
    file_path = f"{file_path}/{filename}"

    # 保存文件到本地
    with open(file_path, "wb") as buffer:
        shutil.copyfileobj(file.file, buffer)
    settings = get_settings()
    url_suffix = file_path.replace(str(settings.PIPELINE_DIR),"")
    ts_str = str(int(time.time()))

    url = f"/brave-api/pipeline-dir{url_suffix}?v={ts_str}"
    return {"filename": filename, "url":url}



import textwrap
import os
from typing import Any, Optional
from .base_analysis import BaseAnalysis
from brave.api.core.evenet_bus import EventBus
from brave.api.service import pipeline as pipeline_service

class ScriptAnalysis(BaseAnalysis):
    def __init__(self, event_bus:EventBus) -> None:
        super().__init__(event_bus)


    # def _get_query_db_field(self, conn, component):
    #     return ["metaphlan_sam_abundance"]
    
    def _get_command(self,analysis_id,output_dir,cache_dir,params_path,work_dir,executor_log,component_script,trace_file,workflow_log_file,pieline_dir_with_namespace,script_type) -> str:
        output_path = f"{output_dir}/output"
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        
        language = ""
        if  script_type == "python":
            language = "python"
        elif script_type == "r":
            language = "Rscript"
        elif script_type == "shell":
            language = "bash"
        elif script_type == "jupyter":
            language = "ipython"
        command =  textwrap.dedent(f"""
            {language} {component_script} {params_path}  {output_path}
            """)
        return command
    
    def write_config(self,output_dir,component_script):
        script_config_file = f"{output_dir}/main.config"
        return script_config_file





from operator import and_, or_

from shortuuid import uuid
from brave.microbe.schemas.entity import PageEntity
from brave.microbe.models.core import t_study
from sqlalchemy import select,func
from sqlalchemy.engine import Connection
from brave.microbe.service import graph_service

def page_study(conn: Connection, query: PageEntity):
    stmt =select(
        t_study,
    ) 
    
    # Left join with t_namespace to get namespace name
    # stmt = stmt.select_from(
    #    t_pipeline_components.outerjoin(t_namespace, t_pipeline_components.c.namespace == t_namespace.c.namespace_id)
    # )

    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
            or_(
                t_study.c.study_id.ilike(keyword_pattern),
                t_study.c.study_name.ilike(keyword_pattern),
                t_study.c.description.ilike(keyword_pattern)
            )
        )
    

    # stmt = stmt.where(and_(*conditions))
    # count_stmt = select(func.count()).select_from(t_taxonomy).where(and_(*conditions))
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count()).select_from(t_study).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_study)
    if query.page_size != -1:
        stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_study = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    return {
        "items": find_study,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }

def find_study_by_id(conn: Connection, entity_id: str):
    stmt = t_study.select().where(t_study.c.entity_id ==entity_id)
    find_study = conn.execute(stmt).mappings().first()
    return find_study

def update_study(conn: Connection, entity_id: str, updateData: dict):
    stmt = t_study.update().where(t_study.c.entity_id == entity_id).values(updateData)
    conn.execute(stmt)  


def add_study(conn: Connection, data: dict):
    data['entity_id'] =uuid()
    stmt = t_study.insert().values(data)
    conn.execute(stmt)


def find_study_by_keywords(conn: Connection, keywords: str):
    keyword_pattern = f"%{keywords.strip()}%"
    stmt = t_study.select().where(
        t_study.c.entity_name.ilike(keyword_pattern),
        # or_(
        #     t_study.c.entity_name.ilike(keyword_pattern),
        #     t_study.c.study_name.ilike(keyword_pattern),
        #     t_study.c.description.ilike(keyword_pattern)
        # )
    ).limit(10)
    find_study = conn.execute(stmt).mappings().all()
    find_study = [
        {k:v for k,v in item.items() if v is not None}
        for item in find_study 
    ]
    return find_study

def details_study_by_id(conn: Connection, entity_id: str):  
    stmt = t_study.select().where(t_study.c.entity_id == entity_id)
    details_study = conn.execute(stmt).mappings().first()
    return details_study


def import_studies(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0
    # with conn.begin():  # type: Connection
    for i in range(0, len(records), batch_size):
        batch = records[i:i + batch_size]
        for item in batch:
            item['entity_id'] = uuid()
        stmt = t_study.insert()
        conn.execute(stmt, batch)
        inserted += len(batch)
        print(f"Inserted {inserted} records...")

    return {"message": f"导入完成，共导入 {inserted} 行"}

def delete_study_by_id(conn: Connection, entity_id: str):
    stmt = t_study.delete().where(t_study.c.entity_id == entity_id)
    conn.execute(stmt)


def mining_study(conn: Connection, entity_id: str):
    stmt = t_study.select().where(t_study.c.entity_id ==entity_id)
    find_study = conn.execute(stmt).mappings().first()
    return find_study


from enum import Enum


class IngressEvent(str, Enum):
    NEXTFLOW_EXECUTOR_EVENT = "nextflow_executor_event"
    HEARTBEAT = "__heartbeat__"

class WatchFileEvent(str,Enum):
    WORKFLOW_LOG ="workflow_log"
    TRACE_LOG="trace_log"

class WorkflowEvent(str,Enum):
    ON_FLOW_BEGIN="on_flow_begin"
    ON_PROCESS_COMPLETE="on_process_complete"
    ON_FLOW_COMPLETE="on_flow_complete"
    ON_JOB_SUBMITTED="on_job_submitted"
    ON_FILE_PUBLISH="on_file_publish"
    WORKFLOW_CLEANUP="workflow_cleanup"

class AnalysisExecutorEvent(str,Enum):
    ON_ANALYSIS_SUBMITTED="on_analysis_submitted"
    ON_ANALYSIS_COMPLETE="on_analysis_complete"
    ON_CONTAINER_PULLED = "on_container_pulled"
    ON_ANALYSIS_FAILED="on_analysis_failed"
    ON_ANALYSIS_STARTED="on_analysis_started"
    ON_ANALYSIS_STOPED="on_analysis_stoped"

class AnalysisResultEvent(str,Enum):
    ON_ANALYSIS_RESULT_ADD="on_analysis_result_add"
    ON_ANALYSIS_RESULT_UPDATE="on_analysis_result_update"
    ON_ANALYSIS_RESULT_DELETE="on_analysis_result_delete"
    ON_ANALYSIS_RESULT_QUERY="on_analysis_result_query"
    ON_ANALYSIS_RESULT_QUERY_BY_ID="on_analysis_result_query_by_id"
    ON_ANALYSIS_RESULT_QUERY_BY_ANALYSIS_ID="on_analysis_result_query_by_analysis_id"
    ON_ANALYSIS_RESULT_QUERY_BY_COMPONENT_ID="on_analysis_result_query_by_component_id"
    ON_ANALYSIS_RESULT_QUERY_BY_SAMPLE_ID="on_analysis_result_query_by_sample_id"


from ast import Dict, Set
import asyncio
import json
import psutil
from watchfiles import awatch
from datetime import datetime
from brave.api.config.db import get_engine
from brave.api.models.core import analysis
from sqlalchemy import select, update
import logging
from importlib.resources import files
from importlib import import_module
import inspect
from brave.api.service.sse_service import SSESessionService
from collections import defaultdict
from typing import Dict, Set,Any,List
import brave.api.service.analysis_service as analysis_service
import brave.api.service.analysis_result_service as analysis_result_service
import brave.api.service.sample_service as sample_service
from functools import lru_cache
from fastapi import Depends
from brave.api.service.listener_files_service import get_listener_files_service
from brave.api.service.listener_files_service import ListenerFilesService
from brave.api.core.evenet_bus import EventBus
# 创建 logger
logger = logging.getLogger(__name__)

class AnalysisResultParse:
    def __init__(self):
        # self.queue_process = asyncio.Queue()
        self.queue_lock = asyncio.Lock()  # 保证数据库更新和队列操作安全
        # self.check_interval = check_interval  # 检查间隔
        # self.listener_files = self._load_listener_files()
        # self.sse_service = sse_service
        # self.listener_files_service = listener_files_service
        self.analysis_id_to_analysis_result: Dict[str, List[Any]] =defaultdict(list)
        self.analysis_id_to_params: Dict[str, Any] = defaultdict(dict)
        self.analysis_id_list: List[str] = []
        self.remove_analysis_id_list: List[str] = []
        self.change_analysis_id_list = asyncio.Queue()
        # self.lock = asyncio.Lock()
        self.analysis_locks: Dict[str, asyncio.Lock] = defaultdict(asyncio.Lock)

        # self._load_parsed_analysis_result()

 
    async def add_analysis_id(self,analysis_id):
        async with self.analysis_locks[analysis_id]:
            self.analysis_id_list.append(analysis_id)
    async def add_remove_analysis_id(self,analysis_id):
        async with self.analysis_locks[analysis_id]:
            self.remove_analysis_id_list.append(analysis_id)

    async def add_change_analysis_id(self,analysis_id):
        async with self.analysis_locks[analysis_id]:
            if analysis_id  in self.analysis_id_list:
                await self.change_analysis_id_list.put(analysis_id)

 

    def remove_analysis_id(self,analysis_id):
        if analysis_id in self.analysis_id_list:
            self.analysis_id_list.remove(analysis_id)
            # self.reomve_analysis_result(analysis_id)

    def reomve_analysis_result(self,analysis_id):
        if analysis_id in self.analysis_id_to_analysis_result:
            del self.analysis_id_to_analysis_result[analysis_id]
        if analysis_id in self.analysis_id_to_params:
            del self.analysis_id_to_params[analysis_id]

    
    def remove_analysis_result_by_analsyis_result_id(self,analysis_id,analysis_result_id):
        if analysis_id in self.analysis_id_to_analysis_result:
            analysis_result_list = self.analysis_id_to_analysis_result[analysis_id]
            analysis_result_list = [item for item in analysis_result_list if item['analysis_result_id'] != analysis_result_id]
            self.analysis_id_to_analysis_result[analysis_id] = analysis_result_list

    
    # def is_analysis_id_exist(self,analysis_id):
    #     return analysis_id in self.change_analysis_id_list


   
    def cached_analysis_result(self):
        return self.analysis_id_to_analysis_result
    def cached_params(self):
        return self.analysis_id_to_params



    def find_parse_params(self,conn,analysis_id,preview):
        if preview:
            return analysis_service.get_parse_analysis_result_params(conn,analysis_id)
        if analysis_id in self.analysis_id_to_params:
            return self.analysis_id_to_params[analysis_id]
        else:
            params:Any = analysis_service.get_parse_analysis_result_params(conn,analysis_id)
            self.analysis_id_to_params[analysis_id] = params
            return params

    def find_analysis_result_exist(self,conn,analysis_id,component_id,sample_id,project,preview):
        if preview:
            return analysis_result_service.find_analysis_result_exist(conn,component_id,sample_id,project)
        if analysis_id in self.analysis_id_to_analysis_result:
            analysis_result_list = self.analysis_id_to_analysis_result[analysis_id]
            item = next((item for item in analysis_result_list if item['component_id'] == component_id and item['sample_id'] == sample_id and item['project'] == project), None)
            if item:
                return item
            else:
                result = analysis_result_service.find_analysis_result_exist(conn,component_id,sample_id,project)
                if result:
                    self.analysis_id_to_analysis_result[analysis_id].append(result)
                return result
        else:
            result = analysis_result_service.find_analysis_result_exist(conn,component_id,sample_id,project)
            if result:
                self.analysis_id_to_analysis_result[analysis_id].append(result)
            return result
            
    def find_by_sample_name_and_project(self,conn,sample_name,project):
        result = sample_service.find_by_sample_name_and_project(conn,sample_name,project)
        return result
    def find_file_dict(self,file_format_list,output_dir):
        file_dict = analysis_service.get_file_dict(file_format_list,output_dir)
        return file_dict


    async def auto_save_analysis_result(self):
        while True:
            try:
                analysis_id = await asyncio.wait_for(self.change_analysis_id_list.get(), timeout=0.5)
                print(f"自动解析分析: {analysis_id} !")
                with get_engine().begin() as conn:
                    await self.save_analysis_result(conn,analysis_id,True)

            except asyncio.TimeoutError:
                if self.change_analysis_id_list.empty():
                    for analysis_id in self.remove_analysis_id_list:
                        await asyncio.sleep(0.05)
                        # print(f"删除分析: {analysis_id} !")
                        # self.remove_analysis_id(analysis_id)
                        # self.reomve_analysis_result(analysis_id)
                        # self.remove_analysis_id_list = []
           
            # async with self.lock:
            # if not self.is_analysis_id_exist(analysis_id):
            #     return
            
  
            
            

    
    async def save_analysis_result(self,conn,analysis_id,preview):
     
        params,result_list,result_dict = self.parse_analysis_result(conn,analysis_id,preview)
        print("result_list",json.dumps(result_list,indent=4))
        for item in result_list:    
            result = self.find_analysis_result_exist(conn,analysis_id,item['component_id'],item['file_name'],item['project'],preview)
            if not result:
                find_sample = self.find_by_sample_name_and_project(conn,item['file_name'],item['project'])
                if find_sample:
                    item['sample_id'] = find_sample['sample_id']
              
                
                analysis_result_service.add_analysis_result(conn,item)
                # await self.listener_files_service.execute_listener("analysis_result_add",{
                #     "analysis":params['analysis'],
                #     "analysis_result":item,
                #     "sse_service":self.sse_service
                # })
                # data = json.dumps({
                #     "msg":f"分析{analysis_id}，文件{item['file_name']}保存成功!",
                #     "component_id":item['component_id'],
                #     "msgType":"analysis_result"
                # })
                # msg = {"group": "default", "data": data}
                # await self.sse_service.push_message(msg)
            else:
                if item['analysis_result_hash']!= result['analysis_result_hash']:
                    analysis_result_service.update_analysis_result(conn,result.id,item)
                    # await self.listener_files_service.execute_listener("analysis_result_update",{
                    #     "analysis":params['analysis'],
                    #     "analysis_result":item,
                    #     "sse_service":self.sse_service
                    # })
                    # data = json.dumps({
                    #     "msg":f"分析{analysis_id}，文件{item['file_name']}更新成功!",
                    #     "msgType":"analysis_result"
                    # })
                    # msg = {"group": "default", "data": data}
                    # await self.sse_service.push_message(msg)
        return params,result_list,result_dict


    def parse_analysis_result(self,conn,analysis_id,preview):
        params:Any = self.find_parse_params(conn,analysis_id,preview)
   
        result_list,result_dict = analysis_service.execute_parse(**params)
        return params,result_list,result_dict

    async def save_analysis_result_preview(self,conn,analysis_id):
        params,result_list,result_dict = await self.save_analysis_result(conn,analysis_id,True)
        file_format_list = params["file_format_list"]
        file_dict = self.find_file_dict(file_format_list,params['analysis']['output_dir'])   
        return {"result_dict":result_dict,"file_format_list":file_format_list,"file_dict":file_dict}

        
    async def parse_analysis_result_preview(self,conn,analysis_id):
        params,result_list,result_dict = self.parse_analysis_result(conn,analysis_id,True)
        file_format_list = params["file_format_list"]
        file_dict = self.find_file_dict(file_format_list,params['analysis']['output_dir'])   
        return {"result_dict":result_dict,"file_format_list":file_format_list,"file_dict":file_dict}
   



import nbformat as nbf
from nbconvert import PythonExporter



def generate_notebook(path):
    nb = nbf.v4.new_notebook()

    cells = [
        nbf.v4.new_markdown_cell("# 自动生成的 Notebook\n这里写一些说明"),
        nbf.v4.new_code_cell("import sys\nprint(sys.argv)"),
        nbf.v4.new_code_cell("print('Hello from generated notebook!')"),
    ]

    nb['cells'] = cells

    with open(path, "w", encoding="utf-8") as f:
        nbf.write(nb, f)


def convert_notebook(source_path, target_path):  
    # 读取 Notebook
    with open(source_path, "r", encoding="utf-8") as f:
        nb = nbf.read(f, as_version=4)

    # 创建 Python 导出器
    exporter = PythonExporter()
    python_code, _ = exporter.from_notebook_node(nb)

    # 写入 .py 文件
    with open(target_path, "w", encoding="utf-8") as f:
        f.write(python_code)
        print(f"写入文件: {target_path}")


# independent 模型处理器
from brave.api.service.components.base import ComponentHandler


class IndependentComponentHandler(ComponentHandler):
    def create_files(self, component_data):
        print(f"[Independent] 创建独立组件文件: {component_data}")

    def get_files(self, component_data):
        print(f"[Independent] 获取独立组件文件: {component_data}")

    def update_files(self, component_data):
        print(f"[Independent] 更新独立组件文件: {component_data}")

    def delete_files(self, component_data):
        print(f"[Independent] 删除独立组件文件: {component_data}")


import re
import json
from shortuuid import uuid

# 原始文本列表
# lines = [
#     "AMetabolism",
#     "B  Global and overview maps",
#     "C    01100  Metabolic pathways",
#     "C    01110  Biosynthesis of secondary metabolites",
#     "C    01120  Microbial metabolism in diverse environments",
#     "C    01200  Carbon metabolism",
#     "C    01210  2-Oxocarboxylic acid metabolism",
#     "C    01212  Fatty acid metabolism",
#     "C    01230  Biosynthesis of amino acids",
#     "C    01232  Nucleotide metabolism",
#     "C    01250  Biosynthesis of nucleotide sugars",
#     "C    01240  Biosynthesis of cofactors",
#     "C    01220  Degradation of aromatic compounds",
#     "C    01310  Nitrogen cycle",
#     "C    01320  Sulfur cycle"
# ]

category = "KEGG"
last_parent = {}  # 保存每个层级最新的 tree_number
output = []

def generate_entity_id(code):
    return f"map{code}"

def parse_kegg_hierarchy(lines):
    output.append({
        "entity_name": "KEGG",
        "category": "KEGG",
        "entity_id": uuid(),
        "parent_tree": None,
        "tree_number": "KEGG"
    })
    for line in lines:
        line = line.strip()
        if not line:
            continue

        level = line[0]  # A/B/C
        content = line[1:].strip()

        node = {"category": category}

        if level == 'A':
            node["entity_name"] = content  # 根节点
            node["category"] = category
            node["parent_tree"] = category
            children_count = sum(1 for n in output if n.get("parent_tree") == category)
            tree_number = f"{category}.{children_count + 1:03d}"
            last_parent['A'] = tree_number
            node["entity_id"] = uuid()
            
        elif level == 'B':
            node["entity_name"] = content
            parent_tree = last_parent['A']
            node["parent_tree"] = parent_tree
            node["entity_id"] = uuid()
            # 生成 tree_number
            children_count = sum(1 for n in output if n.get("parent_tree") == parent_tree)
            tree_number = f"{parent_tree}.{children_count + 1:03d}"
            last_parent['B'] = tree_number
        elif level == 'C':
            # C 层可能带 code
            m = re.match(r"(\d{5})\s+(.*)", content)
            if m:
                code, name = m.groups()
                node["entity_name"] = name
                node["entity_id"] = generate_entity_id(code)
            else:
                node["entity_name"] = content
            parent_tree = last_parent['B']
            node["parent_tree"] = parent_tree
            children_count = sum(1 for n in output if n.get("parent_tree") == parent_tree)
            tree_number = f"{parent_tree}.{children_count + 1:03d}"
        node["tree_number"] = tree_number
        output.append(node)
    return output

    # 输出 JSON
# print(json.dumps(output, indent=2, ensure_ascii=False))



from abc import ABC, abstractmethod
from ast import Dict
import json
import os
from typing import Any, Optional
from brave.api.config.db import get_engine
from fastapi import HTTPException
from brave.api.core.evenet_bus import EventBus
from brave.api.enum.component_script import ScriptName
from brave.api.schemas.analysis import AnalysisExecuterModal
import  brave.api.service.pipeline as pipeline_service
import  brave.api.service.analysis_result_service as analysis_result_service
import  brave.api.service.bio_database_service as bio_database_service
from  brave.api.models.core import analysis as t_analysis
from brave.api.config.config import get_settings
from sqlalchemy import select
import textwrap
import uuid
import importlib
from brave.api.utils.get_db_utils import get_ids,get_group
from brave.api.core.routers_name import RoutersName
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.service import project_service


class BaseAnalysis(ABC):
    def __init__(self, event_bus:EventBus) -> None:
        self.event_bus = event_bus

    # @abstractmethod
    # def _get_query_db_field(self,conn,component):
    #     pass
    def _get_query_db_field(self,conn,component):
        if component['component_type']=="software":
            component_file_list = pipeline_service.find_component_by_parent_id(conn,component['component_id'],"software_input_file")
            component_file_name_list = [json.loads(item.content)['name'] for item in component_file_list]
            return component_file_name_list
        elif component['component_type'] == "script":
            if "formJson" in component:
                return [item['name'] for item in component['formJson'] if "db" in item and  item['db']]
        elif component['component_type'] == "pipeline":
            input_file_list = component['inputFile']
            return [item['name'] for item in input_file_list]
        return []

    @abstractmethod
    def _get_command(self,analysis_id,output_dir,cache_dir,params_path,work_dir,executor_log,component_script,trace_file,workflow_log_file,pieline_dir_with_namespace,script_type) -> str:
        pass
    
    @abstractmethod
    def write_config(self,output_dir,component_script) -> str:
        pass
    
  



    async def save_analysis(self,conn,request_param,parse_analysis_result,component,is_submit):
        # parse_analysis_result,component = self.get_parames(request_param)


        new_analysis = {
            "project":request_param['project'],
            "analysis_name":request_param['analysis_name'],
            "request_param":json.dumps(request_param),
            # "analysis_method":component_script,
            "component_id":component['component_id'],
            # "is_report":request_param['is_report'] if "is_report" in request_param else False,
            "data_component_ids":request_param['data_component_ids'],
            # "analysis_status": "running" if is_submit else "created"
            # "parse_analysis_module":parse_analysis_module
        }
        new_analysis = {k:v for k,v in new_analysis.items() if v is not None}
        # module_dir = pipeline_id
        # if "moduleDir" in component_content:
        #     module_dir = component_content['moduleDir']

        output_dir=None
        work_dir=None
        result = None
        if "analysis_id" in request_param:
            stmt = select(t_analysis).where(t_analysis.c.analysis_id == request_param['analysis_id'])
            result = conn.execute(stmt).mappings().first()
        if result:
            output_dir = result.output_dir
            work_dir = result.work_dir
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
            if not os.path.exists(work_dir):
                os.makedirs(work_dir)
            params_path = result.params_path
            command_path = result.command_path

            with open(params_path, "w") as f:
                json.dump(parse_analysis_result,f)

            # new_analysis['container_id'] = component["container_id"]
            if result.run_type =="job" and result.analysis_status != "running":
                new_analysis["analysis_status"] = "updated"
            # new_analysis['output_format'] = parse_analysis_result_module
            stmt = t_analysis.update().values(new_analysis).where(t_analysis.c.analysis_id==request_param['analysis_id'])
            conn.execute(stmt)
            find_analysis = dict(result)
            new_analysis ={
                **find_analysis,
                **new_analysis
            }
        else:
            settings = get_settings()
            base_dir = settings.BASE_DIR
            work_dir = settings.WORK_DIR
            pieline_dir = settings.PIPELINE_DIR
            str_uuid = str(uuid.uuid4())
            pieline_dir_with_namespace = f"{pieline_dir}"
            # /ssd1/wy/workspace2/nextflow_workspace
            # wrap_analysis_pipline = ""
            # if 'wrap_analysis_pipeline' in request_param:
            # wrap_analysis_pipline = request_param['wrap_analysis_pipeline']
            

            project_dir = f"{base_dir}/{request_param['project']}"
            trace_file = f"{base_dir}/monitor/{str_uuid}.trace.log"
            workflow_log_file = f"{base_dir}/monitor/{str_uuid}.workflow.log"
            # if "pipeline_id" in request_param:  
            #     pipeline_id = request_param['pipeline_id']
            #     output_dir = f"{project_dir}/{pipeline_id}/{component['component_id']}/{str_uuid}"
            # else:
            output_dir = f"{project_dir}/{component['component_id']}/{str_uuid}"
            # /data/wangyang/nf_work/
            work_dir = f"{work_dir}/{request_param['project']}/{component['component_id']}/{str_uuid}"
            params_path = f"{output_dir}/params.json"
            command_path= f"{output_dir}/run.sh"
            command_log_path= f"{output_dir}/run.log"
            # cache_dir = f"{project_dir}/.nextflow"
            cache_dir = f"{output_dir}/.nextflow"
            executor_log = f"{output_dir}/.nextflow.log"
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
            if not os.path.exists(work_dir):
                os.makedirs(work_dir)
            # 写入脚本
        
            # script_dir = pipeline_id
            # if "scriptDir" in component_content:
            #     script_dir = component_content['scriptDir']
            component_script = pipeline_service.find_component_module(component,ScriptName.main)['path']
            # try:
                

         
            # pipeline_script =  f"{get_pipeline_file(pipeline_script)}"
            new_analysis['pipeline_script'] = component_script
            command = self._get_command(str_uuid,output_dir,cache_dir,params_path,work_dir,executor_log,component_script,trace_file,workflow_log_file,pieline_dir_with_namespace,component["script_type"])

            # command =  textwrap.dedent(f"""
            # export BRAVE_WORKFLOW_ID={str_uuid}
            # export NXF_CACHE_DIR={cache_dir}
            # nextflow -log {executor_log} run -offline -resume  \\
            #     -ansi-log false \\
            #     {component_script} \\
            #     -params-file {params_path} \\
            #     -w {work_dir} \\
            #     -plugins nf-hello@0.7.0 \\
            #     -with-trace {trace_file} | tee {workflow_log_file}
            # """)
    
            
            script_config_file = self.write_config(output_dir,component_script)
            # script_config_file = f"{output_dir}/nextflow.config"
            # script_config =  textwrap.dedent(f"""
            # trace.overwrite = true
          
            # """)
            # with open(script_config_file, "w") as f:
            #     f.write(script_config)

            new_analysis['work_dir'] = work_dir
            new_analysis['output_dir'] = output_dir
            new_analysis['params_path'] = params_path
            new_analysis['command_path'] = command_path
            new_analysis['analysis_id'] = str_uuid
            new_analysis['trace_file'] = trace_file
            new_analysis['workflow_log_file'] = workflow_log_file
            new_analysis['executor_log_file'] = executor_log
            new_analysis['script_config_file'] = script_config_file
            new_analysis['command_log_path'] = command_log_path
            # new_analysis['container_id'] = component["container_id"]
            new_analysis["analysis_status"] = "created"
            with open(command_path, "w") as f:
                f.write(command)
            with open(params_path, "w") as f:
                json.dump(parse_analysis_result,f)
            stmt = t_analysis.insert().values(new_analysis)
            conn.execute(stmt)
        if is_submit:
            await self.submit_analysis(new_analysis)
        return new_analysis

    

    def get_parames(self, conn, request_param: dict[str, Any],component):
         # request_param = analysis_input.model_dump_json()
        # component_id = request_param['component_id']
        # pipeline_id = request_param['pipeline_id']
        # if component_id is None:
        #     raise HTTPException(status_code=500, detail=f"component_id is None")

        # component = pipeline_service.find_pipeline_by_id(conn, component_id)
        # if component is None or not hasattr(component, "content"):
        #     raise HTTPException(status_code=404, detail=f"Component with id {component.component_id} not found or missing content.")
        query_name_list = self._get_query_db_field(conn,component)

        parse_analysis_result = self.parse_analysis(conn,request_param,component,query_name_list)
        return parse_analysis_result
    
        
    
            

            # print()
            # return conn.execute(samples.select()).fetchall()
            # print()
        # return {"msg":"success"}

    def get_database_dict(self,conn,value):
        bio_database = bio_database_service.get_bio_database_by_id(conn,value)
        db_index = bio_database.get("db_index")
        if db_index and db_index!="":
            return {
                "db_index":db_index,
                "path":bio_database.get("path")
            }
        else:
            return {
                "path":bio_database.get("path")
            }

    def parse_analysis(self,conn,request_param,component,query_name_list):

        module_info = pipeline_service.find_component_module(component, ScriptName.input_parse)
        if not module_info:
            raise HTTPException(status_code=500, detail=f"组件{component['component_id']}的输入解析模块没有找到!")
        py_module = module_info['module']
        # module_name = f'brave.api.parse_analysis.{module_name}'
        # if importlib.util.find_spec(module) is None:
        #     print(f"{module_name}不存在!")
        # else:
        module = importlib.import_module(py_module)

        

        ## 查找输入字段
        
        # if hasattr(module,"get_db_field"):
        #     get_db_field = getattr(module, "get_db_field")
        #     db_field = get_db_field()
        db_ids_dict = {key: get_ids(request_param[key]) for key in query_name_list if key in request_param}
        db_dict = { key:analysis_result_service.find_analyais_result_by_ids(conn,value) for key,value in  db_ids_dict.items()}
        project_list = [item["project"] for item_list in db_dict.values() for item in item_list]
        project_list = list(set(project_list))
        metadata_form =[]
        if len(project_list)>0:
            project_list =  project_service.find_by_project_ids(conn,project_list)
            metadata_form = [json.loads(item["metadata_form"]) for item in project_list if item["metadata_form"]]
            metadata_form = [item for item_list in metadata_form for item in item_list if item is not None]

        groups_name = {key:get_group(request_param[key]) for key in query_name_list if key in request_param }
        
        extra_dict={}
        if "upstreamFormJson" in component:
            upstream_form_json = component['upstreamFormJson']
            upstream_form_json_names = [item['name'] for item in upstream_form_json]
            extra_dict = {key: request_param[key] for key in upstream_form_json_names if key in request_param}


        if "formJson" in component:
            form_json = component['formJson']
            form_json_names = [item['name'] for item in form_json]
            extra_dict = {key: request_param[key] for key in form_json_names if key in request_param}

        
        database_dict={}
        if "databases" in component:
            bio_database = component['databases']
            bio_database_data_type_list = [item['name'] for item in bio_database]
            db_ids_dict = {key: request_param[key] for key in bio_database_data_type_list if key in request_param}
            database_dict = { key:self.get_database_dict(conn,value) for key,value in  db_ids_dict.items()}

        settings = get_settings()
        args = {
           
            "database_dict":database_dict,
            "extra_dict":extra_dict,
            "analysis_dict":db_dict,
            "groups_name":groups_name,
            "groups":query_name_list,
            "settings":settings,
            "metadata_form":metadata_form
        }

        parse_data = getattr(module, "parse_data")

        result = parse_data(**args)
        return result


    def change_status(self,conn,analysis):
        stmt = t_analysis.update().values({"analysis_status":"running"}).where(t_analysis.c.analysis_id==analysis.analysis_id)
        conn.execute(stmt)

    async def submit_analysis(self,analysis):
        analysis = AnalysisExecuterModal(**analysis)
        await self.event_bus.dispatch(RoutersName.ANALYSIS_EXECUTER_ROUTER,AnalysisExecutorEvent.ON_ANALYSIS_SUBMITTED,analysis)
        





from datetime import datetime
from sqlalchemy import Column, DateTime, Table
from sqlalchemy.sql.sqltypes import Integer, String,Boolean
from brave.api.config.db import meta
# from sqlalchemy.dialects.mysql import LONGTEXT
from sqlalchemy import Text,Index,UniqueConstraint
from sqlalchemy.dialects.mysql import LONGTEXT



t_taxonomy = Table(
    "taxonomy",
    meta,
    # --- nodes.dmp 字段 ---
    Column("tax_id", Integer, primary_key=True, autoincrement=False),   

    Column("entity_id", String(255),  nullable=False, index=True),   

    # tax_id: 物种节点ID (唯一编号)，作为主键

    Column("parent_tax_id", Integer, nullable=False, index=True),       
    # parent_tax_id: 父节点ID，用于构建层级关系 (species → genus → family ...)

    Column("rank", String(50), nullable=False, index=True),             
    # rank: 分类等级 (domain, kingdom, phylum, class, order, family, genus, species)

    # Column("embl_code", String(20)),                                    
    # embl_code: EMBL 数据库 locus 前缀（非唯一，可为空）

    Column("division_id", Integer),                              
    # division_id: division.dmp 中定义的 division ID（生物学分区，例如 Bacteria, Vertebrates）
    
    Column("division_name",  String(255)), 
    # Column("division_code", Integer),     
    # Column("inherited_div_flag", SmallInteger),                         
    # inherited_div_flag: 是否继承 division (1 = 是，0 = 否)

    # Column("genetic_code_id", Integer),                                 
    # genetic_code_id: gencode.dmp 中定义的遗传密码表 ID

    # Column("inherited_gc_flag", SmallInteger),                          
    # inherited_gc_flag: 是否继承遗传密码表 (1 = 是，0 = 否)

    # Column("mitochondrial_genetic_code_id", Integer),                   
    # mitochondrial_genetic_code_id: gencode.dmp 中的线粒体遗传密码表 ID

    # Column("inherited_mgc_flag", SmallInteger),                         
    # inherited_mgc_flag: 是否继承线粒体遗传密码表 (1 = 是，0 = 否)

    # Column("genbank_hidden_flag", SmallInteger),                        
    # genbank_hidden_flag: 是否在 GenBank lineage 中隐藏该节点 (1 = 隐藏)

    # Column("hidden_subtree_root_flag", SmallInteger),                   
    # hidden_subtree_root_flag: 是否为“未有序列数据的子树根” (1 = 是)

    # Column("comments", Text),                                           
    # comments: 自由文本注释（引用、备注）

    # --- names.dmp (只保留 scientific name) ---
    Column("entity_name", String(255), nullable=False, index=True),
    # scientific_name: 学名（唯一科学名称，过滤 name_class='scientific name'）

    # 额外索引：提高 scientific_name 的查询效率
    # Index("idx_name", "scientific_name"),
)

# t_entity = Table(
#     "entity",
#     meta,
#     Column("id", Integer, primary_key=True),
#     Column("entity_id", String(255)),
#     Column("entity_name", String(255)),
#     Column("reference_id", String(255)),
#     Column("is_research", Boolean, default=False),
#     Column("created_at", DateTime, default=datetime.now),
#     Column("updated_at", DateTime, onupdate=datetime.now)
# )


t_study = Table(
    "study",
    meta,
    # 主键：每个研究唯一

    Column("entity_id", String(50), primary_key=True, index=True),
    Column("parent_entity_id", String(50),  index=True),       

    Column("entity_name", String(200) , nullable=False, index=True),
    Column("title", Text, comment="研究标题"),
    Column("overall_design", Text, comment="研究总体设计"),
    Column("criteria_for_disorder", Text, comment="疾病诊断或入选标准"),
    Column("fulltext",Text().with_variant(LONGTEXT(), "mysql")),

    Column("sample_size", Integer, comment="单组样本量"),
    Column("total_sample_size", Integer, comment="总样本量"),
    Column("age", String(50), comment="平均年龄或年龄范围"),
    Column("sex", String(50), comment="性别比例，如 male/female"),
    Column("bmi", String(50), comment="BMI，若有"),
    Column("amplicon_region", String(50), comment="16S rRNA测序区域"),
    
    Column("processing_software", String(255), comment="数据处理软件"),
    Column("reference_database", String(255), comment="参考数据库"),
    Column("original_data_available", String(255), comment="原始数据是否可用"),
    
    Column("pmcid", String(255), comment="PubMed ID"),
    Column("pmid", String(255), comment="PubMed ID"),
    Column("doi", String(255), comment="DOI"),
    Column("country", String(255), comment="研究国家/地区"),
    
    Column("citation", Text, comment="完整引用信息"),
    Column("pubmed_link", String(255), comment="PubMed链接"),
    Column("doi_link", String(255), comment="DOI链接"),
)


t_disease = Table(
    "disease",
    meta,
    Column("entity_id", String(50), primary_key=True, index=True),
    Column("entity_name", String(255) , nullable=False, index=True),
    Column("mesh_id", String(255) , nullable=False, index=True),
    Column("parent_entity_id", String(255),  index=True)

)    

# CREATE INDEX idx_taxonomy_entity_id ON taxonomy(entity_id);
# CREATE INDEX idx_mesh_second_entity_id ON mesh(second_entity_id);

# SELECT * from mesh 
# LEFT JOIN mesh_tree on mesh.entity_id = mesh_tree.entity_id
# where category='B01'

# SELECT * from mesh 
# LEFT JOIN mesh_tree on mesh.entity_id = mesh_tree.entity_id
# LEFT  JOIN taxonomy on mesh.second_entity_id =taxonomy.entity_id 
# where category='B01'

t_mesh = Table(
    "mesh",
    meta,
    Column("id", Integer, primary_key=True, autoincrement=True), 
    Column("entity_id", String(50),nullable=False),
    Column("entity_name", String(255) , nullable=False, index=True),
    # Column("second_entity_id", String(50), index=True),
    Column("is_research", Boolean, default=False),
    Column("entity_type", String(255)),
    Column("tags",Text().with_variant(LONGTEXT(), "mysql")),
    Column("short_name",Text().with_variant(LONGTEXT(), "mysql")),
    Column("describe",Text().with_variant(LONGTEXT(), "mysql")),

    Column("entity_name_zh", String(255) , index=True),
    UniqueConstraint("entity_id", name="uniq_entity_id")
    
) 
t_registry_numbers = Table(
    "registry_numbers",
    meta,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("entity_id", String(50), nullable=False, index=True),  # 外键关联到 t_mesh.entity_id
    Column("registry_number", String(50), nullable=False, index=True),
    Column("registry_type", String(50), nullable=False, server_default="_NULL_", index=True),  # CAS / EC / TXID / Other
    UniqueConstraint("entity_id", "registry_number", name="uniq_registry_number")
)
t_mesh_tree  = Table(
    "mesh_tree",
    meta,
    Column("id", Integer, primary_key=True, autoincrement=True), 
    Column("entity_id", String(50) , nullable=False, index=True),
    Column("tree_number", String(255),  index=True),
    Column("category", String(50),  index=True),
    Column("parent_tree", String(255), nullable=False, server_default="_NULL_", index=True),
    UniqueConstraint("entity_id", "tree_number", "parent_tree", name="uniq_mesh_tree")
) 

# Column("major_category", String(50),  index=True),

# t_intevention = Table(
#     "intevention",
#     meta,
#     Column("entity_id", String(50), primary_key=True, index=True),
#     Column("entity_name", String(255) , nullable=False, index=True),
#     Column("parent_entity_id", String(255),  index=True)
# )    

t_chemicals_and_drugs  = Table(
    "chemicals_and_drugs",
    meta,
    Column("entity_id", String(50), primary_key=True, index=True),
    Column("entity_name", String(255) , nullable=False, index=True),
    Column("parent_entity_id", String(255),  index=True)
)   

t_diet_and_food  = Table(
    "diet_and_food",
    meta,
    Column("entity_id", String(50), primary_key=True, index=True),
    Column("entity_name", String(255) , nullable=False, index=True),
    Column("parent_entity_id", String(255),  index=True)
)  

t_association  = Table(
    "association",
    meta,
    Column("id", Integer, primary_key=True, autoincrement=True), 
    Column("entity_id", String(255),  index=True),
    Column("subject_id", String(255)), # Subject = 作用者（who does something）
    Column("object_id", String(255)), # Object = 被作用者（who receives the effect）
    Column("observed_id", String(255)), 
    Column("study_id", String(255)),
    Column("predicate", String(255)),
    Column("effect", String(255)),
    Column("created_at", DateTime, default=datetime.now),
    Column("updated_at", DateTime, onupdate=datetime.now)
    # Column("participates_in_pathway", String(255)),
    # Column("produces_metabolite", String(255)),
    # Column("regulates_gene", String(255)),
)


from fastapi import APIRouter
# import spacy

from brave.api.config.db import get_engine
from brave.microbe.service import study_service
# from simstring.feature_extractor.character_ngram import CharacterNgramFeatureExtractor
# from simstring.feature_extractor.word_ngram import WordNgramFeatureExtractor

# from simstring.measure.cosine import CosineMeasure
# from simstring.database.dict import DictDatabase
# from simstring.searcher import Searcher

nlp_api = APIRouter(prefix="/nlp")

@nlp_api.post("/find-entity/{entity_id}")
async def add_tokens(entity_id):
    # nlp = spacy.load("en_core_web_sm")
    # with get_engine().begin() as conn:
    #     study = study_service.find_study_by_id(conn,entity_id)
    #     content = study["fulltext"]
    #     doc = nlp(content)
    #     sentences = list(doc.sents)
    #     # tokens =[
    #     #     [token.text for token in sent]
    #     #     for sent in sentences
    #     # ]
    #     sent =[
    #         sent.text
    #         for sent in sentences
    #     ]

        pass

@nlp_api.post("/init-db")
async def init_db():
    pass
    # db = DictDatabase(CharacterNgramFeatureExtractor(2))
    # db.add('foo')
    # db.add('bar')
    # db.add('fooo')
    # # db.save("simstring.db")
    # searcher = Searcher(db, CosineMeasure())
    # results = searcher.search('aa foo nn', 0.8)
    # print(results)


# 抽象基类
class ComponentHandler:
    def create_files(self, component_data):
        raise NotImplementedError
    def update_files(self, component_data):
        raise NotImplementedError
    def delete_files(self, component_data):
        raise NotImplementedError
    def get_files(self, component_data):
        raise NotImplementedError

    
